# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import tracing
from activitysim.core import config
from activitysim.core import pipeline
from activitysim.core import simulate
from activitysim.core import inject
from activitysim.core import expressions

from .util import estimation

logger = logging.getLogger(__name__)


@inject.step()
def free_parking(
        persons_merged, persons,
        chunk_size, trace_hh_id):
    """

    """

    trace_label = 'free_parking'
    model_settings_file_name = 'free_parking.yaml'

    choosers = persons_merged.to_frame()
    choosers = choosers[choosers.workplace_zone_id > -1]
    logger.info("Running %s with %d persons", trace_label, len(choosers))

    model_settings = config.read_model_settings(model_settings_file_name)
    estimator = estimation.manager.begin_estimation('free_parking')

    constants = config.get_model_constants(model_settings)

    # - preprocessor
    preprocessor_settings = model_settings.get('preprocessor', None)
    if preprocessor_settings:

        locals_d = {}
        if constants is not None:
            locals_d.update(constants)

        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_d,
            trace_label=trace_label)

    model_spec = simulate.read_model_spec(file_name=model_settings['SPEC'])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df)
        estimator.write_choosers(choosers)

    choices = simulate.simple_simulate(
        choosers=choosers,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name='free_parking_at_work',
        estimator=estimator)

    free_parking_alt = model_settings['FREE_PARKING_ALT']
    choices = (choices == free_parking_alt)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, 'persons', 'free_parking_at_work')
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    persons = persons.to_frame()
    persons['free_parking_at_work'] = choices.reindex(persons.index).fillna(0).astype(bool)

    pipeline.replace_table("persons", persons)

    tracing.print_summary('free_parking', persons.free_parking_at_work, value_counts=True)

    if trace_hh_id:
        tracing.trace_df(persons,
                         label=trace_label,
                         warn_if_empty=True)



import logging

import numpy as np
import pandas as pd

from activitysim.core import chunk
from activitysim.core import config
from activitysim.core import expressions
from activitysim.core import inject
from activitysim.core import logit
from activitysim.core import pipeline
from activitysim.core import simulate
from activitysim.core import tracing

from activitysim.abm.models.util.trip import get_time_windows
from activitysim.core.interaction_sample_simulate import eval_interaction_utilities
from activitysim.core.simulate import set_skim_wrapper_targets
from activitysim.core.util import reindex


logger = logging.getLogger(__name__)

MAIN_LEG_DURATION = 'main_leg_duration'
IB_DURATION = 'inbound_duration'
OB_DURATION = 'outbound_duration'

TOUR_ID = 'tour_id'
TRIP_ID = 'trip_id'
TOUR_LEG_ID = 'tour_leg_id'
PATTERN_ID = 'pattern_id'
TRIP_DURATION = 'trip_duration'
STOP_TIME_DURATION = 'stop_time_duration'
TRIP_NUM = 'trip_num'
TRIP_COUNT = 'trip_count'
OUTBOUND = 'outbound'

MAX_TOUR_ID = int(1e9)


def generate_tour_leg_id(tour_leg_row):
    return tour_leg_row.tour_id + (int(MAX_TOUR_ID) if tour_leg_row.outbound else int(2 * MAX_TOUR_ID))


def get_tour_legs(trips):
    tour_legs = trips.groupby([TOUR_ID, OUTBOUND], as_index=False)[TRIP_NUM].max()
    tour_legs[TOUR_LEG_ID] = tour_legs.apply(generate_tour_leg_id, axis=1)
    tour_legs = tour_legs.set_index(TOUR_LEG_ID)
    return tour_legs


def trip_departure_rpc(chunk_size, choosers, trace_label):

    # NOTE we chunk chunk_id
    num_choosers = choosers['chunk_id'].max() + 1

    chooser_row_size = choosers.shape[1] + 1

    # scale row_size by average number of chooser rows per chunk_id
    rows_per_chunk_id = choosers.shape[0] / num_choosers
    row_size = (rows_per_chunk_id * chooser_row_size)

    return chunk.rows_per_chunk(chunk_size, row_size, num_choosers, trace_label)


def generate_alternatives(trips, alternative_col_name):
    """
    This method creates an alternatives list of all possible
    trip durations less than the total trip leg duration. If
    the trip only has one trip on the leg, the trip alternative
    only has one alternative for that trip equal to the trip
    duration.
    :param trips: pd.DataFrame
    :param alternative_col_name: column name for the alternative column
    :return: pd.DataFrame
    """
    legs = trips[trips[TRIP_COUNT] > 1]

    leg_alts = None
    durations = np.where(legs[OUTBOUND], legs[OB_DURATION], legs[IB_DURATION])
    if len(durations) > 0:
        leg_alts = pd.Series(np.concatenate([np.arange(0, duration + 1) for duration in durations]),
                             np.repeat(legs.index, durations + 1),
                             name=alternative_col_name).to_frame()

    single_trips = trips[trips[TRIP_COUNT] == 1]
    single_alts = None
    durations = np.where(single_trips[OUTBOUND], single_trips[OB_DURATION], single_trips[IB_DURATION])
    if len(durations) > 0:
        single_alts = pd.Series(durations, single_trips.index,
                                name=alternative_col_name).to_frame()

    if not legs.empty and not single_trips.empty:
        return pd.concat([leg_alts, single_alts])

    return leg_alts if not legs.empty else single_alts


def build_patterns(trips, time_windows):
    tours = trips.groupby([TOUR_ID])[[TRIP_DURATION, TRIP_COUNT]].first()
    duration_and_counts = tours[[TRIP_DURATION, TRIP_COUNT]].values

    # We subtract 1 here, because we already know
    # the one trip of the tour leg based on main tour
    # leg duration
    max_trip_count = trips[TRIP_COUNT].max() - 1

    patterns = []
    pattern_sizes = []

    for duration, trip_count in duration_and_counts:
        possible_windows = time_windows[:trip_count-1, np.where(time_windows[:trip_count-1].sum(axis=0) == duration)[0]]
        possible_windows = np.unique(possible_windows, axis=1).transpose()
        filler = np.full((possible_windows.shape[0], max_trip_count), np.nan)
        filler[:possible_windows.shape[0], :possible_windows.shape[1]] = possible_windows
        patterns.append(filler)
        pattern_sizes.append(filler.shape[0])

    patterns = np.concatenate(patterns)
    pattern_names = ['_'.join('%0.0f' % x for x in y[~np.isnan(y)]) for y in patterns]
    indexes = np.repeat(tours.index, pattern_sizes)

    # If we've done everything right, the indexes
    # calculated above should be the same length as
    # the pattern options
    assert patterns.shape[0] == len(indexes)

    patterns = pd.DataFrame(index=indexes, data=patterns)
    patterns.index.name = tours.index.name
    patterns[PATTERN_ID] = pattern_names

    patterns = patterns.melt(id_vars=PATTERN_ID, value_name=STOP_TIME_DURATION,
                             var_name=TRIP_NUM, ignore_index=False).reset_index()
    patterns = patterns[~patterns[STOP_TIME_DURATION].isnull()].copy()

    patterns[TRIP_NUM] = patterns[TRIP_NUM] + 1
    patterns[STOP_TIME_DURATION] = patterns[STOP_TIME_DURATION].astype(np.int)

    patterns = pd.merge(patterns, trips.reset_index()[[TOUR_ID, TRIP_ID, TRIP_NUM, OUTBOUND]],
                        on=[TOUR_ID, TRIP_NUM])

    patterns.index = patterns.apply(generate_tour_leg_id, axis=1)
    patterns.index.name = TOUR_LEG_ID

    return patterns


def get_spec_for_segment(omnibus_spec, segment):

    spec = omnibus_spec[[segment]]

    # might as well ignore any spec rows with 0 utility
    spec = spec[spec.iloc[:, 0] != 0]
    assert spec.shape[0] > 0

    return spec


def trip_departure_calc_row_size(choosers, trace_label):
    """
    rows_per_chunk calculator for trip_scheduler
    """

    sizer = chunk.RowSizeEstimator(trace_label)

    chooser_row_size = len(choosers.columns)
    spec_columns = 3

    sizer.add_elements(chooser_row_size + spec_columns, 'choosers')

    row_size = sizer.get_hwm()
    return row_size


def choose_tour_leg_pattern(trip_segment,
                            patterns, spec,
                            trace_label='trace_label'):
    alternatives = generate_alternatives(trip_segment, STOP_TIME_DURATION).sort_index()
    have_trace_targets = tracing.has_trace_targets(trip_segment)

    if have_trace_targets:
        tracing.trace_df(trip_segment, tracing.extend_trace_label(trace_label, 'choosers'))
        tracing.trace_df(alternatives, tracing.extend_trace_label(trace_label, 'alternatives'),
                         transpose=False)

    if len(spec.columns) > 1:
        raise RuntimeError('spec must have only one column')

    # - join choosers and alts
    # in vanilla interaction_simulate interaction_df is cross join of choosers and alternatives
    # interaction_df = logit.interaction_dataset(choosers, alternatives, sample_size)
    # here, alternatives is sparsely repeated once for each (non-dup) sample
    # we expect alternatives to have same index of choosers (but with duplicate index values)
    # so we just need to left join alternatives with choosers
    assert alternatives.index.name == trip_segment.index.name

    interaction_df = alternatives.join(trip_segment, how='left', rsuffix='_chooser')

    chunk.log_df(trace_label, 'interaction_df', interaction_df)

    if have_trace_targets:
        trace_rows, trace_ids = tracing.interaction_trace_rows(interaction_df, trip_segment)

        tracing.trace_df(interaction_df,
                         tracing.extend_trace_label(trace_label, 'interaction_df'),
                         transpose=False)
    else:
        trace_rows = trace_ids = None

    interaction_utilities, trace_eval_results \
        = eval_interaction_utilities(spec, interaction_df, None, trace_label, trace_rows, None)

    interaction_utilities = pd.concat([interaction_df[STOP_TIME_DURATION], interaction_utilities], axis=1)
    chunk.log_df(trace_label, 'interaction_utilities', interaction_utilities)

    interaction_utilities = pd.merge(interaction_utilities.reset_index(),
                                     patterns[patterns[TRIP_ID].isin(interaction_utilities.index)],
                                     on=[TRIP_ID, STOP_TIME_DURATION], how='left')

    if have_trace_targets:
        tracing.trace_interaction_eval_results(trace_eval_results, trace_ids,
                                               tracing.extend_trace_label(trace_label, 'eval'))

        tracing.trace_df(interaction_utilities,
                         tracing.extend_trace_label(trace_label, 'interaction_utilities'),
                         transpose=False)

    del interaction_df
    chunk.log_df(trace_label, 'interaction_df', None)

    interaction_utilities = interaction_utilities.groupby([TOUR_ID, OUTBOUND, PATTERN_ID],
                                                          as_index=False)[['utility']].sum()

    interaction_utilities[TOUR_LEG_ID] = \
        interaction_utilities.apply(generate_tour_leg_id, axis=1)

    tour_choosers = interaction_utilities.set_index(TOUR_LEG_ID)
    interaction_utilities = tour_choosers[['utility']].copy()

    # reshape utilities (one utility column and one row per row in model_design)
    # to a dataframe with one row per chooser and one column per alternative
    # interaction_utilities is sparse because duplicate sampled alternatives were dropped
    # so we need to pad with dummy utilities so low that they are never chosen

    # number of samples per chooser
    sample_counts = interaction_utilities.groupby(interaction_utilities.index).size().values
    chunk.log_df(trace_label, 'sample_counts', sample_counts)

    # max number of alternatvies for any chooser
    max_sample_count = sample_counts.max()

    # offsets of the first and last rows of each chooser in sparse interaction_utilities
    last_row_offsets = sample_counts.cumsum()
    first_row_offsets = np.insert(last_row_offsets[:-1], 0, 0)

    # repeat the row offsets once for each dummy utility to insert
    # (we want to insert dummy utilities at the END of the list of alternative utilities)
    # inserts is a list of the indices at which we want to do the insertions
    inserts = np.repeat(last_row_offsets, max_sample_count - sample_counts)

    del sample_counts
    chunk.log_df(trace_label, 'sample_counts', None)

    # insert the zero-prob utilities to pad each alternative set to same size
    padded_utilities = np.insert(interaction_utilities.utility.values, inserts, -999)
    del inserts

    del interaction_utilities
    chunk.log_df(trace_label, 'interaction_utilities', None)

    # reshape to array with one row per chooser, one column per alternative
    padded_utilities = padded_utilities.reshape(-1, max_sample_count)
    chunk.log_df(trace_label, 'padded_utilities', padded_utilities)

    # convert to a dataframe with one row per chooser and one column per alternative
    utilities_df = pd.DataFrame(
        padded_utilities,
        index=tour_choosers.index.unique())
    chunk.log_df(trace_label, 'utilities_df', utilities_df)

    del padded_utilities
    chunk.log_df(trace_label, 'padded_utilities', None)

    if have_trace_targets:
        tracing.trace_df(utilities_df, tracing.extend_trace_label(trace_label, 'utilities'),
                         column_labels=['alternative', 'utility'])

    # convert to probabilities (utilities exponentiated and normalized to probs)
    # probs is same shape as utilities, one row per chooser and one column for alternative
    probs = logit.utils_to_probs(utilities_df,
                                 trace_label=trace_label, trace_choosers=trip_segment)

    chunk.log_df(trace_label, 'probs', probs)

    del utilities_df
    chunk.log_df(trace_label, 'utilities_df', None)

    if have_trace_targets:
        tracing.trace_df(probs, tracing.extend_trace_label(trace_label, 'probs'),
                         column_labels=['alternative', 'probability'])

    # make choices
    # positions is series with the chosen alternative represented as a column index in probs
    # which is an integer between zero and num alternatives in the alternative sample
    positions, rands = \
        logit.make_choices(probs, trace_label=trace_label, trace_choosers=trip_segment)

    chunk.log_df(trace_label, 'positions', positions)
    chunk.log_df(trace_label, 'rands', rands)

    del probs
    chunk.log_df(trace_label, 'probs', None)

    # shouldn't have chosen any of the dummy pad utilities
    assert positions.max() < max_sample_count

    # need to get from an integer offset into the alternative sample to the alternative index
    # that is, we want the index value of the row that is offset by <position> rows into the
    # tranche of this choosers alternatives created by cross join of alternatives and choosers

    # resulting pandas Int64Index has one element per chooser row and is in same order as choosers
    choices = tour_choosers[PATTERN_ID].take(positions + first_row_offsets)

    chunk.log_df(trace_label, 'choices', choices)

    if have_trace_targets:
        tracing.trace_df(choices, tracing.extend_trace_label(trace_label, 'choices'),
                         columns=[None, PATTERN_ID])
        tracing.trace_df(rands, tracing.extend_trace_label(trace_label, 'rands'),
                         columns=[None, 'rand'])

    return choices


def apply_stage_two_model(omnibus_spec, trips, chunk_size, trace_label):

    if not trips.index.is_monotonic:
        trips = trips.sort_index()

    # Assign the duration of the appropriate leg to the trip
    trips[TRIP_DURATION] = np.where(trips[OUTBOUND], trips[OB_DURATION], trips[IB_DURATION])

    trips['depart'] = -1

    # If this is the first outbound trip, the choice is easy, assign the depart time
    # to equal the tour start time.
    trips.loc[(trips['trip_num'] == 1) & (trips[OUTBOUND]), 'depart'] = trips['start']

    # If its the first return leg, it is easy too. Just assign the trip start time to the
    # end time minus the IB duration
    trips.loc[(trips['trip_num'] == 1) & (~trips[OUTBOUND]), 'depart'] = trips['end'] - trips[IB_DURATION]

    # The last leg of the outbound tour needs to begin at the start plus OB duration
    trips.loc[(trips['trip_count'] == trips['trip_num']) & (trips[OUTBOUND]), 'depart'] = \
        trips['start'] + trips[OB_DURATION]

    # The last leg of the inbound tour needs to begin at the end time of the tour
    trips.loc[(trips['trip_count'] == trips['trip_num']) & (~trips[OUTBOUND]), 'depart'] = \
        trips['end']

    # Slice off the remaining trips with an intermediate stops to deal with.
    # Hopefully, with the tricks above we've sliced off a lot of choices.
    # This slice should only include trip numbers greater than 2 since the
    side_trips = trips[(trips['trip_num'] != 1) & (trips['trip_count'] != trips['trip_num'])]

    # No processing needs to be done because we have simple trips / tours
    if side_trips.empty:
        assert trips['depart'].notnull().all
        return trips['depart'].astype(int)

    # Get the potential time windows
    time_windows = get_time_windows(side_trips[TRIP_DURATION].max(), side_trips[TRIP_COUNT].max() - 1)

    row_size = chunk_size and trip_departure_calc_row_size(trips, trace_label)

    trip_list = []

    for i, chooser_chunk, chunk_trace_label in \
            chunk.adaptive_chunked_choosers_by_chunk_id(side_trips, chunk_size, row_size, trace_label):

        for is_outbound, trip_segment in chooser_chunk.groupby(OUTBOUND):
            direction = OUTBOUND if is_outbound else 'inbound'
            spec = get_spec_for_segment(omnibus_spec, direction)
            segment_trace_label = '{}_{}'.format(direction, chunk_trace_label)

            patterns = build_patterns(trip_segment, time_windows)

            choices = choose_tour_leg_pattern(trip_segment,
                                              patterns, spec, trace_label=segment_trace_label)

            choices = pd.merge(choices.reset_index(), patterns.reset_index(),
                               on=[TOUR_LEG_ID, PATTERN_ID], how='left')

            choices = choices[['trip_id', 'stop_time_duration']].copy()

            trip_list.append(choices)

    trip_list = pd.concat(trip_list, sort=True).set_index('trip_id')
    trips['stop_time_duration'] = 0
    trips.update(trip_list)
    trips.loc[trips['trip_num'] == 1, 'stop_time_duration'] = trips['depart']
    trips.sort_values(['tour_id', 'outbound', 'trip_num'])
    trips['stop_time_duration'] = trips.groupby(['tour_id', 'outbound'])['stop_time_duration'].cumsum()
    trips.loc[trips['trip_num'] != trips['trip_count'], 'depart'] = trips['stop_time_duration']
    return trips['depart'].astype(int)


@inject.step()
def trip_departure_choice(
        trips,
        trips_merged,
        skim_dict,
        chunk_size,
        trace_hh_id):

    trace_label = 'trip_departure_choice'
    model_settings = config.read_model_settings('trip_departure_choice.yaml')

    spec = simulate.read_model_spec(file_name=model_settings['SPECIFICATION'])

    trips_merged_df = trips_merged.to_frame()
    # add tour-based chunk_id so we can chunk all trips in tour together
    tour_ids = trips_merged[TOUR_ID].unique()
    trips_merged_df['chunk_id'] = reindex(pd.Series(list(range(len(tour_ids))), tour_ids), trips_merged_df.tour_id)

    max_tour_id = trips_merged[TOUR_ID].max()

    trip_departure_choice.MAX_TOUR_ID = int(np.power(10, np.ceil(np.log10(max_tour_id))))
    locals_d = config.get_model_constants(model_settings).copy()

    preprocessor_settings = model_settings.get('PREPROCESSOR', None)
    tour_legs = get_tour_legs(trips_merged_df)
    pipeline.get_rn_generator().add_channel('tour_legs', tour_legs)

    if preprocessor_settings:
        od_skim = skim_dict.wrap('origin', 'destination')
        do_skim = skim_dict.wrap('destination', 'origin')

        skims = [od_skim, do_skim]

        simulate.set_skim_wrapper_targets(trips_merged_df, skims)

        locals_d.update({
            "od_skims": od_skim,
            "do_skims": do_skim,
        })

        expressions.assign_columns(
            df=trips_merged_df,
            model_settings=preprocessor_settings,
            locals_dict=locals_d,
            trace_label=trace_label)

    choices = apply_stage_two_model(spec, trips_merged_df, chunk_size, trace_label)

    trips_df = trips.to_frame()
    trip_length = len(trips_df)
    trips_df = pd.concat([trips_df, choices], axis=1)
    assert len(trips_df) == trip_length
    assert trips_df[trips_df['depart'].isnull()].empty

    pipeline.replace_table("trips", trips_df)



import sys

from activitysim.cli import CLI
from activitysim.cli import run
from activitysim.cli import create

from activitysim import __version__, __doc__


def main():
    asim = CLI(version=__version__,
               description=__doc__)
    asim.add_subcommand(name='run',
                        args_func=run.add_run_args,
                        exec_func=run.run,
                        description=run.run.__doc__)
    asim.add_subcommand(name='create',
                        args_func=create.add_create_args,
                        exec_func=create.create,
                        description=create.create.__doc__)
    sys.exit(asim.execute())



import numpy as np
import pandas as pd
import pytest

from activitysim.abm.models.util.trip import get_time_windows


@pytest.mark.parametrize("duration, levels, expected",
                         [(24, 3, 2925), (24, 2, 325), (24, 1, 25),
                          (48, 3, 20825), (48, 2, 1225), (48, 1, 49)])
def test_get_time_windows(duration, levels, expected):
    time_windows = get_time_windows(duration, levels)

    if levels == 1:
        assert time_windows.ndim == 1
        assert len(time_windows) == expected
        assert np.sum(time_windows <= duration) == expected
    else:
        assert len(time_windows) == levels
        assert len(time_windows[0]) == expected
        total_duration = np.sum(time_windows, axis=0)
        assert np.sum(total_duration <= duration) == expected

    df = pd.DataFrame(np.transpose(time_windows))
    assert len(df) == len(df.drop_duplicates())



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from . import logit
from . import tracing
from . import chunk
from .simulate import set_skim_wrapper_targets

from activitysim.core.mem import force_garbage_collect
from .interaction_simulate import eval_interaction_utilities

logger = logging.getLogger(__name__)


def _interaction_sample_simulate(
        choosers, alternatives, spec,
        choice_column,
        allow_zero_probs, zero_prob_choice_val,
        want_logsums,
        skims, locals_d,
        trace_label, trace_choice_name,
        estimator):

    """
    Run a MNL simulation in the situation in which alternatives must
    be merged with choosers because there are interaction terms or
    because alternatives are being sampled.

    Parameters are same as for public function interaction_sample_simulate

    spec : dataframe
        one row per spec expression and one col with utility coefficient

    interaction_df : dataframe
        cross join (cartesian product) of choosers with alternatives
        combines columns of choosers and alternatives
        len(df) == len(choosers) * len(alternatives)
        index values (non-unique) are index values from alternatives df

    interaction_utilities : dataframe
        the utility of each alternative is sum of the partial utilities determined by the
        various spec expressions and their corresponding coefficients
        yielding a dataframe  with len(interaction_df) rows and one utility column
        having the same index as interaction_df (non-unique values from alternatives df)

    utilities : dataframe
        dot product of model_design.dot(spec)
        yields utility value for element in the cross product of choosers and alternatives
        this is then reshaped as a dataframe with one row per chooser and one column per alternative

    probs : dataframe
        utilities exponentiated and converted to probabilities
        same shape as utilities, one row per chooser and one column for alternative

    positions : series
        choices among alternatives with the chosen alternative represented
        as the integer index of the selected alternative column in probs

    choices : series
        series with the alternative chosen for each chooser
        the index is same as choosers
        and the series value is the alternative df index of chosen alternative

    Returns
    -------
    if want_logsums is False:

        choices : pandas.Series
            A series where index should match the index of the choosers DataFrame
            and values will match the index of the alternatives DataFrame -
            choices are simulated in the standard Monte Carlo fashion

    if want_logsums is True:

        choices : pandas.DataFrame
            choices['choice'] : same as choices series when logsums is False
            choices['logsum'] : float logsum of choice utilities across alternatives
    """

    # merge of alternatives, choosers on index requires increasing index
    assert choosers.index.is_monotonic_increasing
    assert alternatives.index.is_monotonic_increasing

    # assert choosers.index.equals(alternatives.index[~alternatives.index.duplicated(keep='first')])

    # this is the more general check (not requiring is_monotonic_increasing)
    last_repeat = alternatives.index != np.roll(alternatives.index, -1)
    assert (choosers.shape[0] == 1) or choosers.index.equals(alternatives.index[last_repeat])

    have_trace_targets = tracing.has_trace_targets(choosers)

    if have_trace_targets:
        tracing.trace_df(choosers, tracing.extend_trace_label(trace_label, 'choosers'))
        tracing.trace_df(alternatives, tracing.extend_trace_label(trace_label, 'alternatives'),
                         transpose=False)

    if len(spec.columns) > 1:
        raise RuntimeError('spec must have only one column')

    # if using skims, copy index into the dataframe, so it will be
    # available as the "destination" for the skims dereference below
    if skims is not None:
        alternatives[alternatives.index.name] = alternatives.index

    # - join choosers and alts
    # in vanilla interaction_simulate interaction_df is cross join of choosers and alternatives
    # interaction_df = logit.interaction_dataset(choosers, alternatives, sample_size)
    # here, alternatives is sparsely repeated once for each (non-dup) sample
    # we expect alternatives to have same index of choosers (but with duplicate index values)
    # so we just need to left join alternatives with choosers
    assert alternatives.index.name == choosers.index.name

    interaction_df = alternatives.join(choosers, how='left', rsuffix='_chooser')

    chunk.log_df(trace_label, 'interaction_df', interaction_df)

    if have_trace_targets:
        trace_rows, trace_ids = tracing.interaction_trace_rows(interaction_df, choosers)

        tracing.trace_df(interaction_df,
                         tracing.extend_trace_label(trace_label, 'interaction_df'),
                         transpose=False)
    else:
        trace_rows = trace_ids = None

    if skims is not None:
        set_skim_wrapper_targets(interaction_df, skims)

    # evaluate expressions from the spec multiply by coefficients and sum
    # spec is df with one row per spec expression and one col with utility coefficient
    # column names of choosers match spec index values
    # utilities has utility value for element in the cross product of choosers and alternatives
    # interaction_utilities is a df with one utility column and one row per row in alternative
    interaction_utilities, trace_eval_results \
        = eval_interaction_utilities(spec, interaction_df, locals_d, trace_label, trace_rows, estimator)
    chunk.log_df(trace_label, 'interaction_utilities', interaction_utilities)

    del interaction_df
    chunk.log_df(trace_label, 'interaction_df', None)

    if have_trace_targets:
        tracing.trace_interaction_eval_results(trace_eval_results, trace_ids,
                                               tracing.extend_trace_label(trace_label, 'eval'))

        tracing.trace_df(interaction_utilities,
                         tracing.extend_trace_label(trace_label, 'interaction_utilities'),
                         transpose=False)

    # reshape utilities (one utility column and one row per row in model_design)
    # to a dataframe with one row per chooser and one column per alternative
    # interaction_utilities is sparse because duplicate sampled alternatives were dropped
    # so we need to pad with dummy utilities so low that they are never chosen

    # number of samples per chooser
    sample_counts = interaction_utilities.groupby(interaction_utilities.index).size().values
    chunk.log_df(trace_label, 'sample_counts', sample_counts)

    # max number of alternatvies for any chooser
    max_sample_count = sample_counts.max()

    # offsets of the first and last rows of each chooser in sparse interaction_utilities
    last_row_offsets = sample_counts.cumsum()
    first_row_offsets = np.insert(last_row_offsets[:-1], 0, 0)

    # repeat the row offsets once for each dummy utility to insert
    # (we want to insert dummy utilities at the END of the list of alternative utilities)
    # inserts is a list of the indices at which we want to do the insertions
    inserts = np.repeat(last_row_offsets, max_sample_count - sample_counts)

    del sample_counts
    chunk.log_df(trace_label, 'sample_counts', None)

    # insert the zero-prob utilities to pad each alternative set to same size
    padded_utilities = np.insert(interaction_utilities.utility.values, inserts, -999)
    chunk.log_df(trace_label, 'padded_utilities', padded_utilities)
    del inserts

    del interaction_utilities
    chunk.log_df(trace_label, 'interaction_utilities', None)

    # reshape to array with one row per chooser, one column per alternative
    padded_utilities = padded_utilities.reshape(-1, max_sample_count)

    # convert to a dataframe with one row per chooser and one column per alternative
    utilities_df = pd.DataFrame(
        padded_utilities,
        index=choosers.index)
    chunk.log_df(trace_label, 'utilities_df', utilities_df)

    del padded_utilities
    chunk.log_df(trace_label, 'padded_utilities', None)

    if have_trace_targets:
        tracing.trace_df(utilities_df, tracing.extend_trace_label(trace_label, 'utilities'),
                         column_labels=['alternative', 'utility'])

    # convert to probabilities (utilities exponentiated and normalized to probs)
    # probs is same shape as utilities, one row per chooser and one column for alternative
    probs = logit.utils_to_probs(utilities_df, allow_zero_probs=allow_zero_probs,
                                 trace_label=trace_label, trace_choosers=choosers)
    chunk.log_df(trace_label, 'probs', probs)

    if want_logsums:
        logsums = logit.utils_to_logsums(utilities_df)
        chunk.log_df(trace_label, 'logsums', logsums)

    del utilities_df
    chunk.log_df(trace_label, 'utilities_df', None)

    if have_trace_targets:
        tracing.trace_df(probs, tracing.extend_trace_label(trace_label, 'probs'),
                         column_labels=['alternative', 'probability'])

    if allow_zero_probs:
        zero_probs = (probs.sum(axis=1) == 0)
        if zero_probs.any():
            # FIXME this is kind of gnarly, but we force choice of first alt
            probs.loc[zero_probs, 0] = 1.0

    # make choices
    # positions is series with the chosen alternative represented as a column index in probs
    # which is an integer between zero and num alternatives in the alternative sample
    positions, rands = \
        logit.make_choices(probs, trace_label=trace_label, trace_choosers=choosers)

    chunk.log_df(trace_label, 'positions', positions)
    chunk.log_df(trace_label, 'rands', rands)

    del probs
    chunk.log_df(trace_label, 'probs', None)

    # shouldn't have chosen any of the dummy pad utilities
    assert positions.max() < max_sample_count

    # need to get from an integer offset into the alternative sample to the alternative index
    # that is, we want the index value of the row that is offset by <position> rows into the
    # tranche of this choosers alternatives created by cross join of alternatives and choosers

    # resulting pandas Int64Index has one element per chooser row and is in same order as choosers
    choices = alternatives[choice_column].take(positions + first_row_offsets)

    # create a series with index from choosers and the index of the chosen alternative
    choices = pd.Series(choices, index=choosers.index)

    chunk.log_df(trace_label, 'choices', choices)

    if allow_zero_probs and zero_probs.any():
        # FIXME this is kind of gnarly, patch choice for zero_probs
        choices.loc[zero_probs] = zero_prob_choice_val

    if have_trace_targets:
        tracing.trace_df(choices, tracing.extend_trace_label(trace_label, 'choices'),
                         columns=[None, trace_choice_name])
        tracing.trace_df(rands, tracing.extend_trace_label(trace_label, 'rands'),
                         columns=[None, 'rand'])
        if want_logsums:
            tracing.trace_df(logsums, tracing.extend_trace_label(trace_label, 'logsum'),
                             columns=[None, 'logsum'])

    if want_logsums:
        choices = choices.to_frame('choice')
        choices['logsum'] = logsums

    return choices


def interaction_sample_simulate_calc_row_size(choosers, alt_sample, spec, trace_label):

    # It is hard to estimate the size of the utilities_df since it conflates duplicate picks.
    # Currently we ignore it, but maybe we should chunk based on worst case?

    sizer = chunk.RowSizeEstimator(trace_label)

    num_choosers = len(choosers.index)
    chooser_row_size = len(choosers.columns)

    # one column per alternative plus skims, interaction_utilities, probs
    alt_row_size = alt_sample.shape[1] + 3
    # average sample size
    sample_size = alt_sample.shape[0] / float(num_choosers)

    # interaction_df
    sizer.add_elements((chooser_row_size + alt_row_size) * sample_size, 'interaction_df')

    row_size = sizer.get_hwm()
    return row_size


def interaction_sample_simulate(
        choosers, alternatives, spec, choice_column,
        allow_zero_probs=False, zero_prob_choice_val=None,
        want_logsums=False,
        skims=None, locals_d=None, chunk_size=0,
        trace_label=None, trace_choice_name=None,
        estimator=None):

    """
    Run a simulation in the situation in which alternatives must
    be merged with choosers because there are interaction terms or
    because alternatives are being sampled.

    optionally (if chunk_size > 0) iterates over choosers in chunk_size chunks

    Parameters
    ----------
    choosers : pandas.DataFrame
        DataFrame of choosers
    alternatives : pandas.DataFrame
        DataFrame of alternatives - will be merged with choosers
        index domain same as choosers, but repeated for each alternative
    spec : pandas.DataFrame
        A Pandas DataFrame that gives the specification of the variables to
        compute and the coefficients for each variable.
        Variable specifications must be in the table index and the
        table should have only one column of coefficients.
    skims : Skims object
        The skims object is used to contain multiple matrices of
        origin-destination impedances.  Make sure to also add it to the
        locals_d below in order to access it in expressions.  The *only* job
        of this method in regards to skims is to call set_df with the
        dataframe that comes back from interacting choosers with
        alternatives.  See the skims module for more documentation on how
        the skims object is intended to be used.
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    chunk_size : int
        if chunk_size > 0 iterates over choosers in chunk_size chunks
    trace_label: str
        This is the label to be used  for trace log file entries and dump file names
        when household tracing enabled. No tracing occurs if label is empty or None.
    trace_choice_name: str
        This is the column label to be used in trace file csv dump of choices

    Returns
    -------
    if want_logsums is False:

        choices : pandas.Series
            A series where index should match the index of the choosers DataFrame
            and values will match the index of the alternatives DataFrame -
            choices are simulated in the standard Monte Carlo fashion

    if want_logsums is True:

        choices : pandas.DataFrame
            choices['choice'] : same as choices series when logsums is False
            choices['logsum'] : float logsum of choice utilities across alternatives

    """

    trace_label = tracing.extend_trace_label(trace_label, 'interaction_sample_simulate')

    row_size = chunk_size and interaction_sample_simulate_calc_row_size(choosers, alternatives, spec, trace_label)

    result_list = []
    for i, chooser_chunk, alternative_chunk, chunk_trace_label \
            in chunk.adaptive_chunked_choosers_and_alts(choosers, alternatives,
                                                        chunk_size, row_size, trace_label):

        choices = _interaction_sample_simulate(
            chooser_chunk, alternative_chunk, spec, choice_column,
            allow_zero_probs, zero_prob_choice_val, want_logsums,
            skims, locals_d,
            chunk_trace_label, trace_choice_name,
            estimator)

        result_list.append(choices)

        force_garbage_collect()

    # FIXME: this will require 2X RAM
    # if necessary, could append to hdf5 store on disk:
    # http://pandas.pydata.org/pandas-docs/stable/io.html#id2
    if len(result_list) > 1:
        choices = pd.concat(result_list)

    assert len(choices.index == len(choosers.index))

    return choices



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd
import numpy as np

from activitysim.core import tracing
from activitysim.core import inject


logger = logging.getLogger(__name__)


def rle(a):
    """
    Compute run lengths of values in rows of a two dimensional ndarry of ints.

    We assume the first and last columns are buffer columns
    (because this is the case for time windows)  and so don't include them in results.


    Return arrays giving row_id, start_pos, run_length, and value of each run of any length.

    Parameters
    ----------
    a : numpy.ndarray of int shape(n, <num_time_periods_in_a_day>)


        The input array would normally only have values of 0 or 1 to detect overlapping
        time period availability but we don't assume this, and will detect and report
        runs of any values. (Might prove useful in future?...)

    Returns
    -------
    row_id : numpy.ndarray int shape(<num_runs>)
    start_pos : numpy.ndarray int shape(<num_runs>)
    run_length : numpy.ndarray int shape(<num_runs>)
    run_val : numpy.ndarray int shape(<num_runs>)
    """

    # note timewindows have a beginning and end of day padding columns that we must ignore
    # a = [[1, 1, 1, 1, 0, 1, 1, 1, 0, 1],
    #      [1, 1, 1, 0, 0, 1, 1, 0, 1, 1],
    #      [1, 0, 1, 0, 0, 1, 1, 0, 1, 1],
    #      [1, 1, 1, 1, 1, 1, 1, 0, 1, 1]])

    a = np.asarray(a)

    # row element is different from the one before
    changed = np.array(a[..., :-1] != a[..., 1:])

    # first and last real columns always considered different from padding
    changed[..., 0] = True
    changed[..., -1] = True

    # array([[ True, False, False,  True,  True, False, False,  True,  True],
    #        [ True, False,  True, False,  True, False,  True,  True,  True],
    #        [ True,  True,  True, False,  True, False,  True,  True,  True],
    #        [ True, False, False, False, False, False,  True,  True,  True]])

    # indices of change points (row_index, col_index)
    i = np.where(changed)
    # ([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3],
    #  [0, 3, 4, 7, 8, 0, 2, 4, 6, 7, 8, 0, 1, 2, 4, 6, 7, 8, 0, 6, 7, 8])

    row_id = i[0][1:]
    row_changed, run_length = np.diff(i)
    # row_id      [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3]
    # row_changed [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0]
    # run_length  [3, 1, 3, 1,-8, 2, 2, 2, 1, 1,-8, 1, 1, 2, 2, 1, 1,-8, 6, 1, 1]

    # start position of run in changed array + 1 to get pos in input array
    start_pos = np.cumsum(np.append(0, run_length))[:-1] + 1
    # start_pos   [1, 4, 5, 8, 9, 1, 3, 5, 7, 8, 9, 1, 2, 3, 5, 7, 8, 9, 1, 7, 8]

    # drop bogus negative run length when row changes, we want to drop them
    real_rows = np.where(1 - row_changed)[0]
    row_id = row_id[real_rows]
    run_length = run_length[real_rows]
    start_pos = start_pos[real_rows]

    # index into original array to get run_val
    run_val = a[(row_id, start_pos)]

    # real_rows  [0, 1, 2, 3, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15, 16, 18, 19, 20]
    # row_id     [0, 0, 0, 0, 1, 1, 1, 1, 1,  2,  2,  2,  2,  2,  2,  3,  3,  3]
    # run_length [3, 1, 3, 1, 2, 2, 2, 1, 1,  1,  1,  2,  2,  1,  1,  6,  1,  1]
    # start_pos  [1, 4, 5, 8, 1, 3, 5, 7, 8,  1,  2,  3,  5,  7,  8,  1,  7,  8]
    # run_val    [1, 0, 1, 0, 1, 0, 1, 0, 1,  0,  1,  0,  1,  0,  1,  1,  0,  1]

    return row_id, start_pos, run_length, run_val


def p2p_time_window_overlap(p1_ids, p2_ids):
    """

    Parameters
    ----------
    p1_ids
    p2_ids

    Returns
    -------

    """

    timetable = inject.get_injectable("timetable")

    assert len(p1_ids) == len(p2_ids)
    # if series, ought to have same index
    assert (p1_ids.index == p2_ids.index).all()

    # ndarray with one row per p2p and one column per time period
    # array value of 1 where overlapping free periods and 0 elsewhere
    available = timetable.pairwise_available(p1_ids, p2_ids)

    row_ids, start_pos, run_length, run_val = rle(available)

    # rle returns all runs, but we only care about runs of available (run_val == 1)
    target_rows = np.where(run_val == 1)
    row_ids = row_ids[target_rows]
    run_length = run_length[target_rows]

    df = pd.DataFrame({'row_ids': row_ids, 'run_length': run_length})

    # groupby index of row_ids match the numpy row indexes of timetable.pairwise_available ndarray
    # but there may be missing values of any no-overlap persons pairs
    max_overlap = df.groupby('row_ids').run_length.max()
    # fill in any missing values to align with input arrays
    input_row_ids = np.arange(len(p1_ids))
    max_overlap = max_overlap.reindex(input_row_ids).fillna(0)

    # FIXME should we return series or ndarray?
    max_overlap.index = p1_ids.index

    return max_overlap


def person_pairs(persons):

    p = persons[['household_id', 'adult']].reset_index()
    p2p = pd.merge(p, p, left_on='household_id', right_on='household_id', how='outer')

    # we desire well known non-contingent column names
    p2p.rename(columns={
        '%s_x' % persons.index.name: 'person1',
        '%s_y' % persons.index.name: 'person2',
    }, inplace=True)

    p2p = p2p[p2p.person1 < p2p.person2]

    # index is meaningless, but might as well be tidy
    p2p.reset_index(drop=True, inplace=True)

    p2p['p2p_type'] = (p2p.adult_x * 1 + p2p.adult_y * 1).map({0: 'cc', 1: 'ac', 2: 'aa'})

    p2p = p2p[['household_id', 'person1', 'person2', 'p2p_type']]

    return p2p


def hh_time_window_overlap(households, persons):

    p2p = person_pairs(persons)

    p2p['max_overlap'] = p2p_time_window_overlap(p2p.person1, p2p.person2)

    hh_overlap = \
        p2p.groupby(['household_id', 'p2p_type']).max_overlap.max().unstack(level=-1, fill_value=0)

    # fill in missing households (in case there were no overlaps)
    hh_overlap = hh_overlap.reindex(households.index).fillna(0).astype(np.int8)

    # make sure we have all p2p_types (if there were none to unstack, then column will be missing)
    for c in ['aa', 'cc', 'ac']:
        if c not in hh_overlap.columns:
            hh_overlap[c] = 0

    return hh_overlap


def person_time_window_overlap(persons):

    p2p = person_pairs(persons)

    p2p['max_overlap'] = p2p_time_window_overlap(p2p.person1, p2p.person2)

    p_overlap = pd.concat([
        p2p[['person1', 'p2p_type', 'max_overlap']].rename(columns={'person1': 'person_id'}),
        p2p[['person2', 'p2p_type', 'max_overlap']].rename(columns={'person2': 'person_id'})
    ]).groupby(['person_id', 'p2p_type']).max_overlap.max()

    # unstack to create columns for each p2p_type (aa, cc, and ac)
    p_overlap = p_overlap.unstack(level=-1, fill_value=0)

    # make sure we have columns for all p2p_types (in case there were none of a p2ptype to unstack)
    for c in ['aa', 'cc', 'ac']:
        if c not in p_overlap.columns:
            p_overlap[c] = 0

    # fill in missing households (in case there were persons with no overlaps)
    p_overlap = p_overlap.reindex(persons.index).fillna(0).astype(np.int8)

    return p_overlap


def person_max_window(persons):

    timetable = inject.get_injectable("timetable")

    # ndarray with one row per person and one column per time period
    # array value of 1 where free periods and 0 elsewhere
    s = pd.Series(persons.index.values, index=persons.index)
    available = timetable.individually_available(s)

    row_ids, start_pos, run_length, run_val = rle(available)

    # rle returns all runs, but we only care about runs of available (run_val == 1)
    target_rows = np.where(run_val == 1)
    row_ids = row_ids[target_rows]
    run_length = run_length[target_rows]

    df = pd.DataFrame({'row_ids': row_ids, 'run_length': run_length})

    # groupby index of row_ids match the numpy row indexes of timetable.pairwise_available ndarray
    # but there may be missing values of any no-overlap persons pairs
    max_overlap = df.groupby('row_ids').run_length.max()
    # fill in any missing values to align with input arrays
    input_row_ids = np.arange(persons.shape[0])
    max_window = max_overlap.reindex(input_row_ids).fillna(0)

    # FIXME should we return series or ndarray?
    max_window.index = persons.index

    return max_window



# ActivitySim
# See full license in LICENSE.txt.

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from ..util import reindex
from ..util import other_than
from ..util import quick_loc_series
from ..util import quick_loc_df


@pytest.fixture(scope='module')
def people():
    return pd.DataFrame({
        'household': [1, 2, 2, 3, 3, 3, 4, 4, 4, 4],
        'ptype':     [1, 2, 1, 3, 1, 2, 3, 2, 2, 1]},
        index=['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'])


def test_other_than(people):
    expected = pd.Series(
        [False, False, True, True, True, False, True, True, True, True],
        index=people.index, name='left')

    bools = people['ptype'] == 2
    others = other_than(people['household'], bools)

    pdt.assert_series_equal(others, expected)


def test_reindex():
    s = pd.Series([.5, 1.0, 1.5], index=[2, 1, 3])
    s2 = pd.Series([1, 2, 3], index=['a', 'b', 'c'])
    assert list(reindex(s, s2).values) == [1.0, .5, 1.5]


def test_quick_loc_df():

    df = pd.DataFrame({'attrib': ['1', '2', '3', '4', '5']}, index=[1, 2, 3, 4, 5])

    loc_list = np.asanyarray([2, 1, 3, 4, 4, 5, 1])
    attrib_list = [str(i) for i in loc_list]

    assert list(quick_loc_df(loc_list, df, 'attrib')) == attrib_list
    assert list(quick_loc_df(loc_list, df, 'attrib')) == list(df.loc[loc_list]['attrib'])


def test_quick_loc_series():

    series = pd.Series(['1', '2', '3', '4', '5'], index=[1, 2, 3, 4, 5])

    loc_list = np.asanyarray([2, 1, 3, 4, 4, 5, 1])
    attrib_list = [str(i) for i in loc_list]

    assert list(quick_loc_series(loc_list, series)) == attrib_list
    assert list(quick_loc_series(loc_list, series)) == list(series.loc[loc_list])



# ActivitySim
# See full license in LICENSE.txt.
from builtins import input

import math
import logging
from collections import OrderedDict
from contextlib import contextmanager

import numpy as np
import pandas as pd

from . import util
from . import mem
from . import tracing

logger = logging.getLogger(__name__)

# dict of table_dicts keyed by trace_label
# table_dicts are dicts tuples of {table_name: (elements, bytes, mem), ...}
CHUNK_LOG = OrderedDict()

# array of chunk_size active CHUNK_LOG
CHUNK_SIZE = []

HWM = [{}]

INITIAL_ROWS_PER_CHUNK = 10
MAX_ROWSIZE_ERROR = 0.5  # estimated_row_size percentage error warning threshold
INTERACTIVE_TRACE_CHUNKING = False
INTERACTIVE_TRACE_CHUNK_WARNING = False

CHUNK_RSS = False
BYTES_PER_ELEMENT = 8

# chunk size being a bit opaque, it may be helpful to know the chunk size of a small sample run to titrate chunk_size
CHUNK_HISTORY = True  # always log chunk history even if chunk_size == 0


def GB(bytes):
    # symbols = ('', 'K', 'M', 'G', 'T')
    symbols = ('', ' KB', ' MB', ' GB', ' TB')
    fmt = "%.1f%s"
    for i, s in enumerate(symbols):
        units = 1 << i * 10
        if bytes < units * 1024:
            return fmt % (bytes / units, s)


def commas(x):
    x = int(x)
    if x < 10000:
        return str(x)
    result = ''
    while x >= 1000:
        x, r = divmod(x, 1000)
        result = ",%03d%s" % (r, result)
    return "%d%s" % (x, result)


class RowSizeEstimator(object):
    """
    Utility for estimating row_size
    """
    def __init__(self, trace_label):
        self.row_size = 0
        self.hwm = 0  # element count at high water mark
        self.hwm_tag = None  # tag that drove hwm (with ref count)
        self.trace_label = trace_label
        self.elements = {}
        self.tag_count = {}

    def add_elements(self, elements, tag):
        self.elements[tag] = elements
        self.tag_count[tag] = self.tag_count.setdefault(tag, 0) + 1  # number of times tag has been seen
        self.row_size += elements
        logger.debug(f"{self.trace_label} #chunk_calc {tag} {elements} ({self.row_size})")
        input("add_elements>") if INTERACTIVE_TRACE_CHUNKING else None

        if self.row_size > self.hwm:
            self.hwm = self.row_size
            # tag that drove hwm (with ref count)
            self.hwm_tag = f'{tag}_{self.tag_count[tag]}' if self.tag_count[tag] > 1 else tag

    def drop_elements(self, tag):
        self.row_size -= self.elements[tag]
        self.elements[tag] = 0
        logger.debug(f"{self.trace_label} #chunk_calc {tag} <drop> ({self.row_size})")

    def get_hwm(self):
        logger.debug(f"{self.trace_label} #chunk_calc hwm {self.hwm} after {self.hwm_tag}")
        input("get_hwm>") if INTERACTIVE_TRACE_CHUNKING else None
        return self.hwm


def get_high_water_mark(tag='elements'):

    # should always have at least the base chunker
    assert len(HWM) > 0

    hwm = HWM[-1]

    # hwm might be empty if there were no calls to log_df
    mark = hwm.get(tag).get('mark') if hwm else 0

    return mark


@contextmanager
def chunk_log(trace_label, chunk_size=0):
    log_open(trace_label, chunk_size)
    try:
        yield
    finally:
        log_close(trace_label)


def log_open(trace_label, chunk_size=0):

    # nested chunkers should be unchunked
    if len(CHUNK_LOG) > 0:
        assert chunk_size == 0
        assert trace_label not in CHUNK_LOG

    CHUNK_LOG[trace_label] = OrderedDict()
    CHUNK_SIZE.append(chunk_size)

    HWM.append({})


def log_close(trace_label):

    # they should be closing the last log opened (LIFO)
    assert CHUNK_LOG and next(reversed(CHUNK_LOG)) == trace_label

    # if we are closing base level chunker
    if len(CHUNK_LOG) == 1:
        log_write_hwm()

    label, _ = CHUNK_LOG.popitem(last=True)
    assert label == trace_label
    CHUNK_SIZE.pop()
    HWM.pop()


def log_df(trace_label, table_name, df):
    """
    Parameters
    ----------
    trace_label :
        serves as a label for this nesting level of logging
    table_name : str
        name to use logging df, and which will be used in any subsequent calls reporting activity on the table
    df: numpy.ndarray, pandas.Series, pandas.DataFrame, or None
        table to log (or None if df was deleted)
    """

    if df is None:
        # FIXME force_garbage_collect on delete?
        mem.force_garbage_collect()

    try:
        cur_chunker = next(reversed(CHUNK_LOG))
    except StopIteration:
        logger.warning(f"log_df called without current chunker. Did you forget to call log_open?")
        return

    if df is None:
        # remove table from log
        CHUNK_LOG.get(cur_chunker).pop(table_name)
        op = 'del'
        logger.debug(f"log_df {table_name} "
                     f"elements: {0} : {trace_label}")
    else:

        elements = util.iprod(df.shape)
        op = 'add'

        if isinstance(df, pd.Series):
            bytes = df.memory_usage(index=True)
        elif isinstance(df, pd.DataFrame):
            bytes = df.memory_usage(index=True).sum()
        elif isinstance(df, np.ndarray):
            bytes = df.nbytes
        else:
            logger.error("log_df %s unknown type: %s" % (table_name, type(df)))
            assert False

        CHUNK_LOG.get(cur_chunker)[table_name] = (elements, bytes)

        # log this df
        logger.debug(f"log_df {table_name} "
                     f"elements: {commas(elements)} "
                     f"bytes: {GB(bytes)} "
                     f"shape: {df.shape} : {trace_label}")

    hwm_trace_label = "%s.%s.%s" % (trace_label, op, table_name)
    mem.trace_memory_info(hwm_trace_label)

    total_elements, total_bytes = _chunk_totals()  # new chunk totals
    cur_rss = mem.get_rss()

    # - check high_water_marks
    info = f"elements: {commas(total_elements)} " \
           f"bytes: {GB(total_bytes)} " \
           f"rss: {GB(cur_rss)} " \
           f"chunk_size: {commas(CHUNK_SIZE[0])}"

    if INTERACTIVE_TRACE_CHUNKING:
        print(f"table_name {table_name} {df.shape if df is not None else 0}")
        print(f"table_name {table_name} {info}")
        input("log_df>")

    check_for_hwm('elements', total_elements, info, hwm_trace_label)
    check_for_hwm('bytes', total_bytes, info, hwm_trace_label)
    check_for_hwm('rss', cur_rss, info, hwm_trace_label)


def _chunk_totals():

    total_elements = 0
    total_bytes = 0
    for label in CHUNK_LOG:
        tables = CHUNK_LOG[label]
        for table_name in tables:
            elements, bytes = tables[table_name]
            total_elements += elements
            total_bytes += bytes

    return total_elements, total_bytes


def check_for_hwm(tag, value, info, trace_label):

    for d in HWM:

        hwm = d.setdefault(tag, {})

        if value > hwm.get('mark', 0):
            hwm['mark'] = value
            hwm['info'] = info
            hwm['trace_label'] = trace_label


def log_write_hwm():

    d = HWM[0]
    for tag in d:
        hwm = d[tag]
        logger.debug("#chunk_hwm high_water_mark %s: %s (%s) in %s" %
                     (tag, hwm['mark'], hwm['info'], hwm['trace_label']), )

    # - elements shouldn't exceed chunk_size of base chunker
    def check_chunk_size(hwm, chunk_size, label, max_leeway):
        elements = hwm['mark']
        if chunk_size and max_leeway and elements > chunk_size * max_leeway:  # too high
            # FIXME check for #warning in log - there is nothing the user can do about this
            logger.debug("#chunk_hwm #warning total_elements (%s) > %s (%s) %s : %s " %
                         (commas(elements), label, commas(chunk_size),
                          hwm['info'], hwm['trace_label']))

    # if we are in a chunker
    if len(HWM) > 1 and HWM[1]:
        assert 'elements' in HWM[1]  # expect an 'elements' hwm dict for base chunker
        hwm = HWM[1].get('elements')
        check_chunk_size(hwm, CHUNK_SIZE[0], 'chunk_size', max_leeway=1)


def write_history(caller, history, trace_label):

    observed_size = history.observed_chunk_size.sum()
    number_of_rows = history.rows_per_chunk.sum()
    observed_row_size = math.ceil(observed_size / number_of_rows)  # FIXME

    num_chunks = len(history)

    logger.info(f"#chunk_history {caller} {trace_label} "
                f"number_of_rows: {number_of_rows} "
                f"observed_row_size: {observed_row_size} "
                f"num_chunks: {num_chunks}")

    logger.debug(f"#chunk_history {caller} {trace_label}\n{history}")

    initial_row_size = history.row_size.values[0]
    if initial_row_size > 0:

        # if they provided an initial estimated row size, then report error

        error = (initial_row_size - observed_row_size) / observed_row_size
        percent_error = round(error * 100, 1)

        logger.info(f"#chunk_history {caller} {trace_label} "
                    f"initial_row_size: {initial_row_size} "
                    f"observed_row_size: {observed_row_size} "
                    f"percent_error: {percent_error}%")

        if abs(error) > MAX_ROWSIZE_ERROR:

            logger.warning(f"#chunk_history MAX_ROWSIZE_ERROR "
                           f"initial_row_size {initial_row_size} "
                           f"observed_row_size {observed_row_size} "
                           f"percent_error: {percent_error}% in {trace_label}")

            if INTERACTIVE_TRACE_CHUNK_WARNING:
                # for debugging adaptive chunking internals
                print(history)
                input(f"{trace_label} type any key to continue")


def adaptive_chunked_choosers(choosers, chunk_size, row_size, trace_label):

    # generator to iterate over choosers

    num_choosers = len(choosers.index)
    assert num_choosers > 0
    assert chunk_size >= 0
    assert row_size >= 0

    logger.info(f"Running adaptive_chunked_choosers with chunk_size {chunk_size} and {num_choosers} choosers")

    # FIXME do we care if it is an int?
    row_size = math.ceil(row_size)

    # #CHUNK_RSS
    mem.force_garbage_collect()
    initial_rss = mem.get_rss()
    logger.debug(f"#CHUNK_RSS initial_rss: {initial_rss}")

    if chunk_size == 0:
        assert row_size == 0  # we ignore this but make sure caller realizes that
        rows_per_chunk = num_choosers
        estimated_number_of_chunks = 1
    else:
        assert len(HWM) == 1, f"len(HWM): {len(HWM)}"
        if row_size == 0:
            rows_per_chunk = min(num_choosers, INITIAL_ROWS_PER_CHUNK)  # FIXME parameterize
            estimated_number_of_chunks = None
        else:
            max_rows_per_chunk = np.maximum(int(chunk_size / row_size), 1)
            logger.debug(f"#chunk_calc chunk: max rows_per_chunk {max_rows_per_chunk} based on row_size {row_size}")

            rows_per_chunk = np.clip(int(chunk_size / row_size), 1, num_choosers)
            estimated_number_of_chunks = math.ceil(num_choosers / rows_per_chunk)

        logger.debug(f"#chunk_calc chunk: initial rows_per_chunk {rows_per_chunk} based on row_size {row_size}")

    history = {}
    i = offset = 0
    while offset < num_choosers:

        assert offset + rows_per_chunk <= num_choosers

        chunk_trace_label = \
            tracing.extend_trace_label(trace_label, f'chunk_{i + 1}') if chunk_size > 0 else trace_label

        # grab the next chunk based on current rows_per_chunk
        chooser_chunk = choosers.iloc[offset: offset + rows_per_chunk]

        logger.info(f"Running chunk {i+1} of {estimated_number_of_chunks or '?'} "
                    f"with {len(chooser_chunk)} of {num_choosers} choosers")

        with chunk_log(trace_label, chunk_size):

            yield i+1, chooser_chunk, chunk_trace_label

            # get number of elements allocated during this chunk from the high water mark dict
            observed_chunk_size = get_high_water_mark()

            # #CHUNK_RSS
            observed_rss_size = (get_high_water_mark('rss') - initial_rss)
            observed_rss_size = math.ceil(observed_rss_size / BYTES_PER_ELEMENT)
            logger.debug(f"#CHUNK_RSS chunk {i+1} observed_chunk_size: {observed_chunk_size} "
                         f"observed_rss_size {observed_rss_size}")
            observed_rss_size = max(observed_rss_size, 0)
            if CHUNK_RSS:
                observed_chunk_size = observed_rss_size

        i += 1
        offset += rows_per_chunk
        rows_remaining = num_choosers - offset

        if CHUNK_HISTORY or chunk_size > 0:

            history.setdefault('chunk', []).append(i)
            history.setdefault('row_size', []).append(row_size)
            history.setdefault('rows_per_chunk', []).append(rows_per_chunk)
            history.setdefault('observed_chunk_size', []).append(observed_chunk_size)

            # revise predicted row_size based on observed_chunk_size
            row_size = math.ceil(observed_chunk_size / rows_per_chunk)

            # closest number of chooser rows to achieve chunk_size without exceeding it
            if row_size == 0:
                # they don't appear to have used any memory; increase cautiously in case small sample size was to blame
                if rows_per_chunk > INITIAL_ROWS_PER_CHUNK * 100:
                    rows_per_chunk = rows_remaining
                else:
                    rows_per_chunk = 10 * rows_per_chunk
            else:
                rows_per_chunk = int(chunk_size / row_size)
            rows_per_chunk = np.clip(rows_per_chunk, 1, rows_remaining)

            estimated_number_of_chunks = i + math.ceil(rows_remaining / rows_per_chunk) if rows_remaining else i

            history.setdefault('new_row_size', []).append(row_size)
            history.setdefault('new_rows_per_chunk', []).append(rows_per_chunk)
            history.setdefault('estimated_number_of_chunks', []).append(estimated_number_of_chunks)

    if history:
        history = pd.DataFrame.from_dict(history)
        write_history('adaptive_chunked_choosers', history, trace_label)


def adaptive_chunked_choosers_and_alts(choosers, alternatives, chunk_size, row_size, trace_label):
    """
    generator to iterate over choosers and alternatives in chunk_size chunks

    like chunked_choosers, but also chunks alternatives
    for use with sampled alternatives which will have different alternatives (and numbers of alts)

    There may be up to sample_size (or as few as one) alternatives for each chooser
    because alternatives may have been sampled more than once,  but pick_count for those
    alternatives will always sum to sample_size.

    When we chunk the choosers, we need to take care chunking the alternatives as there are
    varying numbers of them for each chooser. Since alternatives appear in the same order
    as choosers, we can use cumulative pick_counts to identify boundaries of sets of alternatives

    Parameters
    ----------
    choosers
    alternatives : pandas DataFrame
        sample alternatives including pick_count column in same order as choosers
    rows_per_chunk : int

    Yields
    -------
    i : int
        one-based index of current chunk
    num_chunks : int
        total number of chunks that will be yielded
    choosers : pandas DataFrame slice
        chunk of choosers
    alternatives : pandas DataFrame slice
        chunk of alternatives for chooser chunk
    """

    # if not choosers.index.is_monotonic_increasing:
    #     logger.warning('sorting choosers because not monotonic increasing')
    #     choosers = choosers.sort_index()

    num_choosers = len(choosers.index)
    num_alternatives = len(alternatives.index)

    assert num_choosers > 0

    # alternatives index should match choosers (except with duplicate repeating alt rows)
    assert choosers.index.equals(alternatives.index[~alternatives.index.duplicated(keep='first')])

    last_repeat = alternatives.index != np.roll(alternatives.index, -1)

    assert (num_choosers == 1) or choosers.index.equals(alternatives.index[last_repeat])
    assert 'pick_count' in alternatives.columns or choosers.index.name == alternatives.index.name
    assert choosers.index.name == alternatives.index.name

    logger.info(f"Running adaptive_chunked_choosers_and_alts with chunk_size {chunk_size} "
                f"and {num_choosers} choosers and {num_alternatives} alternatives")

    # FIXME do we care if it is an int?
    row_size = math.ceil(row_size)

    # #CHUNK_RSS
    mem.force_garbage_collect()
    initial_rss = mem.get_rss()
    logger.debug(f"#CHUNK_RSS initial_rss: {initial_rss}")

    if chunk_size == 0:
        assert row_size == 0  # we ignore this but make sure caller realizes that
        rows_per_chunk = num_choosers
        estimated_number_of_chunks = 1
        row_size = 0
    else:
        assert len(HWM) == 1
        if row_size == 0:
            rows_per_chunk = min(num_choosers, INITIAL_ROWS_PER_CHUNK)  # FIXME parameterize
            estimated_number_of_chunks = None
        else:
            max_rows_per_chunk = np.maximum(int(chunk_size / row_size), 1)
            logger.debug(f"#chunk_calc chunk: max rows_per_chunk {max_rows_per_chunk} based on row_size {row_size}")

            rows_per_chunk = np.clip(int(chunk_size / row_size), 1, num_choosers)
            estimated_number_of_chunks = math.ceil(num_choosers / rows_per_chunk)
        logger.debug(f"#chunk_calc chunk: initial rows_per_chunk {rows_per_chunk} based on row_size {row_size}")

    assert (rows_per_chunk > 0) and (rows_per_chunk <= num_choosers)

    # alt chunks boundaries are where index changes
    alt_ids = alternatives.index.values
    alt_chunk_ends = np.where(alt_ids[:-1] != alt_ids[1:])[0] + 1
    alt_chunk_ends = np.append([0], alt_chunk_ends)  # including the first to simplify indexing
    alt_chunk_ends = np.append(alt_chunk_ends, [len(alternatives.index)])  # end of final chunk

    history = {}
    i = offset = alt_offset = 0
    while offset < num_choosers:
        assert offset + rows_per_chunk <= num_choosers

        chunk_trace_label = tracing.extend_trace_label(trace_label, f'chunk_{i + 1}') if chunk_size > 0 else trace_label

        chooser_chunk = choosers[offset: offset + rows_per_chunk]

        alt_end = alt_chunk_ends[offset + rows_per_chunk]
        alternative_chunk = alternatives[alt_offset: alt_end]

        assert len(chooser_chunk.index) == len(np.unique(alternative_chunk.index.values))
        assert (chooser_chunk.index == np.unique(alternative_chunk.index.values)).all()

        logger.info(f"Running chunk {i+1} of {estimated_number_of_chunks or '?'} "
                    f"with {len(chooser_chunk)} of {num_choosers} choosers")

        with chunk_log(trace_label, chunk_size):

            yield i+1, chooser_chunk, alternative_chunk, chunk_trace_label

            # get number of elements allocated during this chunk from the high water mark dict
            observed_chunk_size = get_high_water_mark()

            # #CHUNK_RSS
            observed_rss_size = (get_high_water_mark('rss') - initial_rss)
            observed_rss_size = math.ceil(observed_rss_size / BYTES_PER_ELEMENT)
            logger.debug(f"#CHUNK_RSS observed_chunk_size: {observed_chunk_size} observed_rss_size {observed_rss_size}")
            observed_rss_size = max(observed_rss_size, 0)
            if CHUNK_RSS:
                observed_chunk_size = observed_rss_size

        alt_offset = alt_end

        i += 1
        offset += rows_per_chunk
        rows_remaining = num_choosers - offset

        if CHUNK_HISTORY or chunk_size > 0:

            history.setdefault('chunk', []).append(i)
            history.setdefault('row_size', []).append(row_size)
            history.setdefault('rows_per_chunk', []).append(rows_per_chunk)
            history.setdefault('observed_chunk_size', []).append(observed_chunk_size)

            # revise predicted row_size based on observed_chunk_size
            row_size = math.ceil(observed_chunk_size / rows_per_chunk)

            # closest number of chooser rows to achieve chunk_size without exceeding it
            if row_size == 0:
                # they don't appear to have used any memory; increase cautiously in case small sample size was to blame
                if rows_per_chunk > INITIAL_ROWS_PER_CHUNK * 100:
                    rows_per_chunk = rows_remaining
                else:
                    rows_per_chunk = 10 * rows_per_chunk
            else:
                rows_per_chunk = int(chunk_size / row_size)
            rows_per_chunk = np.clip(rows_per_chunk, 1, rows_remaining)

            estimated_number_of_chunks = i + math.ceil(rows_remaining / rows_per_chunk) if rows_remaining else i

            history.setdefault('new_row_size', []).append(row_size)
            history.setdefault('new_rows_per_chunk', []).append(rows_per_chunk)
            history.setdefault('estimated_number_of_chunks', []).append(estimated_number_of_chunks)

    if history:
        history = pd.DataFrame.from_dict(history)
        write_history('adaptive_chunked_choosers_and_alts', history, trace_label)


def adaptive_chunked_choosers_by_chunk_id(choosers, chunk_size, row_size, trace_label):
    # generator to iterate over choosers in chunk_size chunks
    # like chunked_choosers but based on chunk_id field rather than dataframe length
    # (the presumption is that choosers has multiple rows with the same chunk_id that
    # all have to be included in the same chunk)
    # FIXME - we pathologically know name of chunk_id col in households table

    num_choosers = choosers['chunk_id'].max() + 1
    assert num_choosers > 0

    # FIXME do we care if it is an int?
    row_size = math.ceil(row_size)

    # #CHUNK_RSS
    mem.force_garbage_collect()
    initial_rss = mem.get_rss()
    logger.debug(f"#CHUNK_RSS initial_rss: {initial_rss}")

    if chunk_size == 0:
        assert row_size == 0  # we ignore this but make sure caller realizes that
        rows_per_chunk = num_choosers
        estimated_number_of_chunks = 1
    else:
        assert len(HWM) == 1

        if row_size == 0:
            rows_per_chunk = min(num_choosers, INITIAL_ROWS_PER_CHUNK)  # FIXME parameterize
            estimated_number_of_chunks = None
            logger.debug(f"#chunk_calc chunk: initial rows_per_chunk {rows_per_chunk} "
                         f"based on INITIAL_ROWS_PER_CHUNK {INITIAL_ROWS_PER_CHUNK}")

        else:
            max_rpc = np.maximum(int(chunk_size / row_size), 1)
            logger.debug(f"#chunk_calc chunk: max rows_per_chunk {max_rpc} based on row_size {row_size}")

            rows_per_chunk = np.clip(int(chunk_size / row_size), 1, num_choosers)
            estimated_number_of_chunks = math.ceil(num_choosers / rows_per_chunk)

            logger.debug(f"#chunk_calc chunk: initial rows_per_chunk {rows_per_chunk} based on row_size {row_size}")

    history = {}
    i = offset = 0
    while offset < num_choosers:

        assert offset + rows_per_chunk <= num_choosers

        chunk_trace_label = \
            tracing.extend_trace_label(trace_label, f'chunk_{i + 1}') if chunk_size > 0 else trace_label

        chooser_chunk = choosers[choosers['chunk_id'].between(offset, offset + rows_per_chunk - 1)]

        logger.info(f"Running chunk {i+1} of {estimated_number_of_chunks or '?'} "
                    f"with {rows_per_chunk} of {num_choosers} choosers")

        with chunk_log(trace_label, chunk_size):

            yield i+1, chooser_chunk, chunk_trace_label

            # get number of elements allocated during this chunk from the high water mark dict
            observed_chunk_size = get_high_water_mark()

            # #CHUNK_RSS
            observed_rss_size = (get_high_water_mark('rss') - initial_rss)
            observed_rss_size = math.ceil(observed_rss_size / BYTES_PER_ELEMENT)
            logger.debug(f"#CHUNK_RSS observed_chunk_size: {observed_chunk_size} observed_rss_size {observed_rss_size}")
            observed_rss_size = max(observed_rss_size, 0)
            if CHUNK_RSS:
                observed_chunk_size = observed_rss_size

        i += 1
        offset += rows_per_chunk
        rows_remaining = num_choosers - offset

        if CHUNK_HISTORY or chunk_size > 0:

            history.setdefault('chunk', []).append(i)
            history.setdefault('row_size', []).append(row_size)
            history.setdefault('rows_per_chunk', []).append(rows_per_chunk)
            history.setdefault('observed_chunk_size', []).append(observed_chunk_size)

            # revise predicted row_size based on observed_chunk_size
            row_size = math.ceil(observed_chunk_size / rows_per_chunk)

            if row_size > 0:
                # closest number of chooser rows to achieve chunk_size without exceeding it
                rows_per_chunk = int(chunk_size / row_size)
            if row_size == 0:
                # they don't appear to have used any memory; increase cautiously in case small sample size was to blame
                if rows_per_chunk > INITIAL_ROWS_PER_CHUNK * 100:
                    rows_per_chunk = rows_remaining
                else:
                    rows_per_chunk = 10 * rows_per_chunk

            rows_per_chunk = np.clip(rows_per_chunk, 1, rows_remaining)

            estimated_number_of_chunks = i + math.ceil(rows_remaining / rows_per_chunk) if rows_remaining else i

            history.setdefault('new_row_size', []).append(row_size)
            history.setdefault('new_rows_per_chunk', []).append(rows_per_chunk)
            history.setdefault('estimated_number_of_chunks', []).append(estimated_number_of_chunks)

    if history:
        history = pd.DataFrame.from_dict(history)
        write_history('adaptive_chunked_choosers_by_chunk_id', history, trace_label)



# ActivitySim
# See full license in LICENSE.txt.
import logging
import warnings
import os
import pandas as pd

from activitysim.core import tracing
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core import expressions
from activitysim.core import mem

from activitysim.core.steps.output import write_data_dictionary
from activitysim.core.steps.output import write_tables
from activitysim.core.steps.output import track_skim_usage

from activitysim.abm.tables import shadow_pricing

# We are using the naming conventions in the mtc_asim.h5 example
# file for our default list. This provides backwards compatibility
# with previous versions of ActivitySim in which only 'input_store'
# is given in the settings file.
DEFAULT_TABLE_LIST = [
    {'tablename': 'households',
     'h5_tablename': 'households',
     'index_col': 'household_id'},
    {'tablename': 'persons',
     'h5_tablename': 'persons',
     'index_col': 'person_id'},
    {'tablename': 'land_use',
     'h5_tablename': 'land_use_taz',
     'index_col': 'TAZ'}
]

logger = logging.getLogger(__name__)


def annotate_tables(model_settings, trace_label):

    annotate_tables = model_settings.get('annotate_tables', [])

    if not annotate_tables:
        logger.warning(f"{trace_label} - annotate_tables setting is empty - nothing to do!")

    assert isinstance(annotate_tables, list), \
        f"annotate_tables settings should be a list but is {type(annotate_tables)}"

    t0 = tracing.print_elapsed_time()

    for table_info in annotate_tables:

        tablename = table_info['tablename']

        df = inject.get_table(tablename).to_frame()

        # - rename columns
        column_map = table_info.get('column_map', None)
        if column_map:

            warnings.warn(f"{trace_label} - annotate_tables option 'column_map' renamed 'rename_columns' "
                          f"and moved to global settings file. Support for 'column_map' in annotate_tables "
                          f"will be removed in future versions.",
                          FutureWarning)

            logger.info(f"{trace_label} - renaming {tablename} columns {column_map}")
            df.rename(columns=column_map, inplace=True)

        # - annotate
        annotate = table_info.get('annotate', None)
        if annotate:
            logger.info(f"{trace_label} - annotating {tablename} SPEC {annotate['SPEC']}")
            expressions.assign_columns(
                df=df,
                model_settings=annotate,
                trace_label=trace_label)

        # fixme - narrow?

        # - write table to pipeline
        pipeline.replace_table(tablename, df)


@inject.step()
def initialize_landuse():

    trace_label = 'initialize_landuse'

    model_settings = config.read_model_settings('initialize_landuse.yaml', mandatory=True)

    annotate_tables(model_settings, trace_label)

    # create accessibility (only required if multiprocessing wants to slice accessibility)
    land_use = pipeline.get_table('land_use')
    accessibility_df = pd.DataFrame(index=land_use.index)
    pipeline.replace_table("accessibility", accessibility_df)


@inject.step()
def initialize_households():

    trace_label = 'initialize_households'

    model_settings = config.read_model_settings('initialize_households.yaml', mandatory=True)
    annotate_tables(model_settings, trace_label)

    # - initialize shadow_pricing size tables after annotating household and person tables
    # since these are scaled to model size, they have to be created while single-process
    shadow_pricing.add_size_tables()
    mem.trace_memory_info(f"initialize_households after shadow_pricing.add_size_tables")

    # - preload person_windows
    t0 = tracing.print_elapsed_time()
    inject.get_table('person_windows').to_frame()
    t0 = tracing.print_elapsed_time("preload person_windows", t0, debug=True)


@inject.injectable(cache=True)
def preload_injectables():
    """
    preload bulky injectables up front - stuff that isn't inserted into the pipeline
    """

    logger.info("preload_injectables")

    inject.add_step('track_skim_usage', track_skim_usage)
    inject.add_step('write_data_dictionary', write_data_dictionary)
    inject.add_step('write_tables', write_tables)

    table_list = config.setting('input_table_list')

    # default ActivitySim table names and indices
    if table_list is None:
        logger.warning(
            "No 'input_table_list' found in settings. This will be a "
            "required setting in upcoming versions of ActivitySim.")

        new_settings = inject.get_injectable('settings')
        new_settings['input_table_list'] = DEFAULT_TABLE_LIST
        inject.add_injectable('settings', new_settings)

    # FIXME undocumented feature
    if config.setting('write_raw_tables'):

        # write raw input tables as csv (before annotation)
        csv_dir = config.output_file_path('raw_tables')
        if not os.path.exists(csv_dir):
            os.makedirs(csv_dir)  # make directory if needed

        table_names = [t['tablename'] for t in table_list]
        for t in table_names:
            df = inject.get_table(t).to_frame()
            if t == 'households':
                df.drop(columns='chunk_id', inplace=True)
            df.to_csv(os.path.join(csv_dir, '%s.csv' % t), index=True)

    t0 = tracing.print_elapsed_time()

    # FIXME - still want to do this?
    # if inject.get_injectable('skim_dict', None) is not None:
    #     t0 = tracing.print_elapsed_time("preload skim_dict", t0, debug=True)
    #
    # if inject.get_injectable('skim_stack', None) is not None:
    #     t0 = tracing.print_elapsed_time("preload skim_stack", t0, debug=True)

    return True



# ActivitySim
# See full license in LICENSE.txt.

import os.path

import numpy.testing as npt
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from .. import inject

from .. import simulate


@pytest.fixture(scope='module')
def data_dir():
    return os.path.join(os.path.dirname(__file__), 'data')


@pytest.fixture(scope='module')
def spec_name(data_dir):
    return 'sample_spec.csv'


@pytest.fixture(scope='module')
def spec(data_dir, spec_name):
    return simulate.read_model_spec(file_name=spec_name)


@pytest.fixture(scope='module')
def data(data_dir):
    return pd.read_csv(os.path.join(data_dir, 'data.csv'))


def setup_function():
    configs_dir = os.path.join(os.path.dirname(__file__), 'configs')
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), f'output')
    inject.add_injectable("output_dir", output_dir)


def test_read_model_spec(spec_name):

    spec = simulate.read_model_spec(file_name=spec_name)

    assert len(spec) == 4
    assert spec.index.name == 'Expression'
    assert list(spec.columns) == ['alt0', 'alt1']
    npt.assert_array_equal(
        spec.values,
        [[1.1, 11], [2.2, 22], [3.3, 33], [4.4, 44]])


def test_eval_variables(spec, data):

    result = simulate.eval_variables(spec.index, data)

    expected = pd.DataFrame([
            [1, 0, 4, 1],
            [0, 1, 4, 1],
            [0, 1, 5, 1]],
            index=data.index, columns=spec.index)

    expected[expected.columns[0]] = expected[expected.columns[0]].astype(np.int8)
    expected[expected.columns[1]] = expected[expected.columns[1]].astype(np.int8)
    expected[expected.columns[2]] = expected[expected.columns[2]].astype(np.int64)
    expected[expected.columns[3]] = expected[expected.columns[3]].astype(int)

    print("\nexpected\n%s" % expected.dtypes)
    print("\nresult\n%s" % result.dtypes)

    pdt.assert_frame_equal(result, expected, check_names=False)


def test_simple_simulate(data, spec):

    inject.add_injectable("settings", {'check_for_variability': False})

    choices = simulate.simple_simulate(choosers=data, spec=spec, nest_spec=None)
    expected = pd.Series([1, 1, 1], index=data.index)
    pdt.assert_series_equal(choices, expected)


def test_simple_simulate_chunked(data, spec):

    inject.add_injectable("settings", {'check_for_variability': False})

    choices = simulate.simple_simulate(choosers=data, spec=spec, nest_spec=None, chunk_size=2)
    expected = pd.Series([1, 1, 1], index=data.index)
    pdt.assert_series_equal(choices, expected)



# ActivitySim
# See full license in LICENSE.txt.
import logging
import itertools
import os

import numpy as np
import pandas as pd
from zbox import toolz as tz, gen

from activitysim.core import simulate
from activitysim.core import pipeline

from activitysim.core import chunk
from activitysim.core import logit
from activitysim.core import tracing
from activitysim.core import inject
from activitysim.core import config

logger = logging.getLogger(__name__)

# FIXME - this allows us to turn some dev debug table dump code on and off - eventually remove?
# DUMP = False

_persons_index_ = 'person_id'
_hh_index_ = 'household_id'
_hh_size_ = 'hhsize'

_hh_id_ = 'household_id'
_ptype_ = 'ptype'
_age_ = 'age'

# For clarity, the named constant MAX_HHSIZE refers to the cdap 5 person threshold figure.
MAX_HHSIZE = 5

MAX_INTERACTION_CARDINALITY = 3


def set_hh_index(df):

    # index on household_id, not person_id
    df.set_index(_hh_id_, inplace=True)
    df.index.name = _hh_index_


def add_pn(col, pnum):
    """
    return the canonical column name for the indiv_util column or columns
    in merged hh_chooser df for individual with cdap_rank pnum

    e.g. M_p1, ptype_p2 but leave _hh_id_ column unchanged
    """
    if type(col) is str:
        return col if col == _hh_id_ else '%s_p%s' % (col, pnum)
    elif isinstance(col, (list, tuple)):
        return [c if c == _hh_id_ else '%s_p%s' % (c, pnum) for c in col]
    else:
        raise RuntimeError("add_pn col not list or str")


def assign_cdap_rank(persons, person_type_map, trace_hh_id=None, trace_label=None):
    """
    Assign an integer index, cdap_rank, to each household member. (Starting with 1, not 0)

    Modifies persons df in place

    The cdap_rank order is important, because cdap only assigns activities to the first
    MAX_HHSIZE persons in each household.

    This will preferentially be two working adults and the three youngest children.

    Rank is assigned starting at 1. This necessitates some care indexing, but is preferred as
    it follows the convention of 1-based pnums in expression files.

    According to the documentation of reOrderPersonsForCdap in mtctm2.abm.ctramp
    HouseholdCoordinatedDailyActivityPatternModel:

    "Method reorders the persons in the household for use with the CDAP model,
    which only explicitly models the interaction of five persons in a HH. Priority
    in the reordering is first given to full time workers (up to two), then to
    part time workers (up to two workers, of any type), then to children (youngest
    to oldest, up to three). If the method is called for a household with less
    than 5 people, the cdapPersonArray is the same as the person array."

    We diverge from the above description in that a cdap_rank is assigned to all persons,
    including 'extra' household members, whose activity is assigned subsequently.
    The pair _hh_id_, cdap_rank will uniquely identify each household member.

    Parameters
    ----------
    persons : pandas.DataFrame
        Table of persons data. Must contain columns _hh_size_, _hh_id_, _ptype_, _age_

    Returns
    -------
    cdap_rank : pandas.Series
        integer cdap_rank of every person, indexed on _persons_index_
    """

    # transient categories used to categorize persons in cdap_rank before assigning final rank
    RANK_WORKER = 1
    RANK_CHILD = 2
    RANK_BACKFILL = 3
    RANK_UNASSIGNED = 9
    persons['cdap_rank'] = RANK_UNASSIGNED

    # choose up to 2 workers, preferring full over part, older over younger
    workers = \
        persons.loc[persons[_ptype_].isin(person_type_map['WORKER']), [_hh_id_, _ptype_]]\
        .sort_values(by=[_hh_id_, _ptype_], ascending=[True, True])\
        .groupby(_hh_id_).head(2)
    # tag the selected workers
    persons.loc[workers.index, 'cdap_rank'] = RANK_WORKER
    del workers

    # choose up to 3, preferring youngest
    children = \
        persons.loc[persons[_ptype_].isin(person_type_map['CHILD']), [_hh_id_, _ptype_, _age_]]\
        .sort_values(by=[_hh_id_, _ptype_], ascending=[True, True])\
        .groupby(_hh_id_).head(3)
    # tag the selected children
    persons.loc[children.index, 'cdap_rank'] = RANK_CHILD
    del children

    # choose up to MAX_HHSIZE, preferring anyone already chosen
    # others = \
    #     persons[[_hh_id_, 'cdap_rank']]\
    #     .sort_values(by=[_hh_id_, 'cdap_rank'], ascending=[True, True])\
    #     .groupby(_hh_id_).head(MAX_HHSIZE)

    # choose up to MAX_HHSIZE, choosing randomly
    others = persons[[_hh_id_, 'cdap_rank']].copy()
    others['random_order'] = pipeline.get_rn_generator().random_for_df(persons)
    others = \
        others\
        .sort_values(by=[_hh_id_, 'random_order'], ascending=[True, True])\
        .groupby(_hh_id_).head(MAX_HHSIZE)

    # tag the backfilled persons
    persons.loc[others[others.cdap_rank == RANK_UNASSIGNED].index, 'cdap_rank'] \
        = RANK_BACKFILL
    del others

    # assign person number in cdapPersonArray preference order
    # i.e. convert cdap_rank from category to index in order of category rank within household
    # groupby rank() is slow, so we compute rank artisanally
    # save time by sorting only the columns we need (persons is big, and sort moves data)
    p = persons[[_hh_id_, 'cdap_rank', _age_]]\
        .sort_values(by=[_hh_id_, 'cdap_rank', _age_], ascending=[True, True, True])
    rank = p.groupby(_hh_id_).size().map(range)
    rank = [item+1 for sublist in rank for item in sublist]
    p['cdap_rank'] = rank
    persons['cdap_rank'] = p['cdap_rank']  # assignment aligns on index values

    # if DUMP:
    #     tracing.trace_df(persons, '%s.DUMP.cdap_person_array' % trace_label,
    #                      transpose=False, slicer='NONE')

    if trace_hh_id:
        tracing.trace_df(persons, '%s.cdap_rank' % trace_label)

    return persons['cdap_rank']


def individual_utilities(
        persons,
        cdap_indiv_spec,
        locals_d,
        trace_hh_id=None, trace_label=None):
    """
    Calculate CDAP utilities for all individuals.

    Parameters
    ----------
    persons : pandas.DataFrame
        DataFrame of individual persons data.
    cdap_indiv_spec : pandas.DataFrame
        CDAP spec applied to individuals.

    Returns
    -------
    utilities : pandas.DataFrame
        Will have index of `persons` and columns for each of the alternatives.
        plus some 'useful columns' [_hh_id_, _ptype_, 'cdap_rank', _hh_size_]

    """

    # calculate single person utilities
    indiv_utils = simulate.eval_utilities(cdap_indiv_spec, persons, locals_d, trace_label=trace_label)

    # add columns from persons to facilitate building household interactions
    useful_columns = [_hh_id_, _ptype_, 'cdap_rank', _hh_size_]
    indiv_utils[useful_columns] = persons[useful_columns]

    if trace_hh_id:
        tracing.trace_df(indiv_utils, '%s.indiv_utils' % trace_label,
                         column_labels=['activity', 'person'])

    return indiv_utils


def preprocess_interaction_coefficients(interaction_coefficients):
    """
    The input cdap_interaction_coefficients.csv file has three columns:

    activity
        A single character activity type name (M, N, or H)

    interaction_ptypes
        List of ptypes in the interaction (in order of increasing ptype)
        Stars `(***)` instead of ptypes means the interaction applies to all ptypes in that size hh.

    coefficient
        The coefficient to apply for all hh interactions for this activity and set of ptypes

    To facilitate building the spec for a given hh ssize, we add two additional columns:

    cardinality
        the number of persons in the interaction (e.g. 3 for a 3-way interaction)

    slug
        a human friendly efficient name so we can dump a readable spec trace file for debugging
        this slug is then replaced with the numerical coefficient value prior to evaluation
    """

    # make a copy
    coefficients = interaction_coefficients.copy()

    if not coefficients['activity'].isin(['M', 'N', 'H']).all():
        msg = "Error in cdap_interaction_coefficients at row %s. Expect only M, N, or H!" \
              % coefficients[~coefficients['activity'].isin(['M', 'N', 'H'])].index.values
        raise RuntimeError(msg)

    coefficients['cardinality'] = coefficients['interaction_ptypes'].astype(str).str.len()

    wildcards = coefficients.interaction_ptypes == coefficients.cardinality.map(lambda x: x*'*')
    coefficients.loc[wildcards, 'interaction_ptypes'] = ''

    coefficients['slug'] = \
        coefficients['activity'] * coefficients['cardinality'] \
        + coefficients['interaction_ptypes'].astype(str)

    return coefficients


def cached_spec_name(hhsize):
    return 'cdap_spec_%s' % hhsize


def get_cached_spec(hhsize):

    spec_name = cached_spec_name(hhsize)

    spec = inject.get_injectable(spec_name, None)
    if spec is not None:
        logger.info("build_cdap_spec returning cached injectable spec %s", spec_name)
        return spec

    # this is problematic for multiprocessing and since we delete csv files in output_dir
    # at the start of every run, doesn't provide any benefit in single-processing as the
    # cached spec will be available as an injectable to subsequent chunks

    # # try data dir
    # if os.path.exists(config.output_file_path(spec_name)):
    #     spec_path = config.output_file_path(spec_name)
    #     logger.info("build_cdap_spec reading cached spec %s from %s", spec_name, spec_path)
    #     return pd.read_csv(spec_path, index_col='Expression')

    return None


def cache_spec(hhsize, spec):
    spec_name = cached_spec_name(hhsize)
    # cache as injectable
    inject.add_injectable(spec_name, spec)


def build_cdap_spec(interaction_coefficients, hhsize,
                    trace_spec=False, trace_label=None, cache=True):
    """
    Build a spec file for computing utilities of alternative household member interaction patterns
    for households of specified size.

    We generate this spec automatically from a table of rules and coefficients because the
    interaction rules are fairly simple and can be expressed compactly whereas
    there is a lot of redundancy between the spec files for different household sizes, as well as
    in the vectorized expression of the interaction alternatives within the spec file itself

    interaction_coefficients has five columns:
        activity
            A single character activity type name (M, N, or H)
        interaction_ptypes
            List of ptypes in the interaction (in order of increasing ptype) or empty for wildcards
            (meaning that the interaction applies to all ptypes in that size hh)
        cardinality
            the number of persons in the interaction (e.g. 3 for a 3-way interaction)
        slug
            a human friendly efficient name so we can dump a readable spec trace file for debugging
            this slug is replaced with the numerical coefficient value after we dump the trace file
        coefficient
            The coefficient to apply for all hh interactions for this activity and set of ptypes

    The generated spec will have the eval expression in the index, and a utility column for each
    alternative (e.g. ['HH', 'HM', 'HN', 'MH', 'MM', 'MN', 'NH', 'NM', 'NN'] for hhsize 2)

    In order to be able to dump the spec in a human-friendly fashion to facilitate debugging the
    cdap_interaction_coefficients table, we first populate utility columns in the spec file
    with the coefficient slugs, dump the spec file, and then replace the slugs with coefficients.

    Parameters
    ----------
    interaction_coefficients : pandas.DataFrame
        Rules and coefficients for generating interaction specs for different household sizes
    hhsize : int
        household size for which the spec should be built.

    Returns
    -------
    spec: pandas.DataFrame

    """

    t0 = tracing.print_elapsed_time()

    # if DUMP:
    #     # dump the interaction_coefficients table because it has been preprocessed
    #     tracing.trace_df(interaction_coefficients,
    #                      '%s.hhsize%d_interaction_coefficients' % (trace_label, hhsize),
    #                      transpose=False, slicer='NONE')

    # cdap spec is same for all households of MAX_HHSIZE and greater
    hhsize = min(hhsize, MAX_HHSIZE)

    if cache:
        spec = get_cached_spec(hhsize)
        if spec is not None:
            return spec

    expression_name = "Expression"

    # generate a list of activity pattern alternatives for this hhsize
    # e.g. ['HH', 'HM', 'HN', 'MH', 'MM', 'MN', 'NH', 'NM', 'NN'] for hhsize=2
    alternatives = [''.join(tup) for tup in itertools.product('HMN', repeat=hhsize)]

    # spec df has expression column plus a column for each alternative
    spec = pd.DataFrame(columns=[expression_name] + alternatives)

    # Before processing the interaction_coefficients, we add add rows to the spec to carry
    # the alternative utilities previously computed for each individual into all hh alternative
    # columns in which the individual assigned that alternative. The Expression column contains
    # the name of the choosers column with that individuals utility for the individual alternative
    # and the hh alternative columns that should receive that utility are given a value of 1
    # e.g. M_p1 is a column in choosers with the individual utility to person p1 of alternative M
    #   Expression   MM   MN   MH   NM   NN   NH   HM   HN   HH
    #         M_p1  1.0  1.0  1.0  0.0  0.0  0.0  0.0  0.0  0.0
    #         N_p1  0.0  0.0  0.0  1.0  1.0  1.0  0.0  0.0  0.0
    for pnum in range(1, hhsize+1):
        for activity in ['M', 'N', 'H']:

            new_row_index = len(spec)
            spec.loc[new_row_index, expression_name] = add_pn(activity, pnum)

            # list of alternative columns where person pnum has expression activity
            # e.g. for M_p1 we want the columns where activity M is in position p1
            alternative_columns = [alt for alt in alternatives if alt[pnum - 1] == activity]
            spec.loc[new_row_index, alternative_columns] = 1

    # ignore rows whose cardinality exceeds hhsize
    relevant_rows = interaction_coefficients.cardinality <= hhsize

    # for each row in the interaction_coefficients table
    for row in interaction_coefficients[relevant_rows].itertuples():

        # if it is a wildcard all_people interaction
        if not row.interaction_ptypes:

            # wildcard interactions only apply if the interaction includes all household members
            # this will be the case if the cardinality of the wildcard equals the hhsize
            # conveniently, the slug is given the name of the alternative column (e.g. HHHH)

            # conveniently, for wildcards, the slug has been assigned the name of the alternative
            # (e.g. HHHH) that it applies to, since the interaction includes all household members
            # and there are no ptypes to append to it

            # FIXME - should we be doing this for greater than HH_MAXSIZE households?
            if row.slug in alternatives:
                spec.loc[len(spec), [expression_name, row.slug]] = ['1', row.slug]

            continue

        if not (0 <= row.cardinality <= MAX_INTERACTION_CARDINALITY):
            raise RuntimeError("Bad row cardinality %d for %s" % (row.cardinality, row.slug))

        # for all other interaction rules, we need to generate a row in the spec for each
        # possible combination of interacting persons
        # e.g. for (1, 2), (1,3), (2,3) for a coefficient with cardinality 2 in hhsize 3
        for tup in itertools.combinations(list(range(1, hhsize+1)), row.cardinality):

            # determine the name of the chooser column with the ptypes for this interaction
            if row.cardinality == 1:
                interaction_column = "ptype_p%d" % tup[0]
            else:
                # column named (e.g.) p1_p3 for an interaction between p1 and p3
                interaction_column = '_'.join(['p%s' % pnum for pnum in tup])

            # build expression that evaluates True iff the interaction is between specified ptypes
            # (e.g.) p1_p3==13 for an interaction between p1 and p3 of ptypes 1 and 3 (or 3 and1 )
            expression = "%s==%s" % (interaction_column, row.interaction_ptypes)

            # create list of columns with names matching activity for each of the persons in tup
            # e.g. ['MMM', 'MMN', 'MMH'] for an interaction between p1 and p3 with activity 'M'
            # alternative_columns = \
            #     filter(lambda alt: all([alt[p - 1] == row.activity for p in tup]), alternatives)
            alternative_columns = \
                [alt for alt in alternatives if all([alt[p - 1] == row.activity for p in tup])]

            # a row for this interaction may already exist,
            # e.g. if there are rules for both HH13 and MM13, we don't need to add rows for both
            # since they are triggered by the same expressions (e.g. p1_p2==13, p1_p3=13,...)
            existing_row_index = (spec[expression_name] == expression)
            if (existing_row_index).any():
                # if the rows exist, simply update the appropriate alternative columns in spec
                spec.loc[existing_row_index, alternative_columns] = row.slug
                spec.loc[existing_row_index, expression_name] = expression
            else:
                # otherwise, add a new row to spec
                new_row_index = len(spec)
                spec.loc[new_row_index, alternative_columns] = row.slug
                spec.loc[new_row_index, expression_name] = expression

    # eval expression goes in the index
    spec.set_index(expression_name, inplace=True)

    simulate.uniquify_spec_index(spec)

    if trace_spec:
        tracing.trace_df(spec, '%s.hhsize%d_spec' % (trace_label, hhsize),
                         transpose=False, slicer='NONE')

    # replace slug with coefficient
    d = interaction_coefficients.set_index('slug')['coefficient'].to_dict()
    for c in spec.columns:
        spec[c] =\
            spec[c].map(lambda x: d.get(x, x or 0.0)).fillna(0)

    if trace_spec:
        tracing.trace_df(spec, '%s.hhsize%d_spec_patched' % (trace_label, hhsize),
                         transpose=False, slicer='NONE')

    if cache:
        cache_spec(hhsize, spec)

    t0 = tracing.print_elapsed_time("build_cdap_spec hh_size %s" % hhsize, t0)

    return spec


def add_interaction_column(choosers, p_tup):
    """
    Add an interaction column in place to choosers, listing the ptypes of the persons in p_tup

    The name of the interaction column will be determined by the cdap_ranks from p_tup,
    and the rows in the column contain the ptypes of those persons in that household row.

    For instance, for p_tup = (1,3) choosers interaction column name will be 'p1_p3'

    For a household where person 1 is part-time worker (ptype=2) and person 3 is infant (ptype 8)
    the corresponding row value interaction code will be 28

    We take advantage of the fact that interactions are symmetrical to simplify spec expressions:
    We name the interaction_column in increasing pnum (cdap_rank) order (p1_p2 and not p3_p1)
    And we format row values in increasing ptype order (28 and not 82)
    This simplifies the spec expressions as we don't have to test for p1_p3 == 28 | p1_p3 == 82

    Parameters
    ----------
    choosers : pandas.DataFrame
        household choosers, indexed on _hh_index_
        choosers should contain columns ptype_p1, ptype_p2 for each cdap_rank person in hh

    p_tup : int tuple
        tuple specifying the cdap_ranks for the interaction column
        p_tup = (1,3) means persons with cdap_rank 1 and 3

    Returns
    -------

    """

    # Since ptypes are always between 1 and 8, we represent the interaction as an integer (24)
    # rather than as a string ('24')
    # FIXME - check that coding interactions as integers is in fact faster then coding as strings

    if p_tup != tuple(sorted(p_tup)):
        raise RuntimeError("add_interaction_column tuple not sorted" % p_tup)

    # FIXME - this could be made more elegant and efficient
    # I couldn't figure out a good way to do this in pandas, but we want to do something like:
    # choosers['p1_p3'] = choosers['ptype_p1'].astype(str) + choosers['ptype_p3'].astype(str)

    dest_col = '_'.join(['p%s' % pnum for pnum in p_tup])

    # build a string concatenating the ptypes of the persons in the order they appear in p_tup
    choosers[dest_col] = choosers[add_pn('ptype', p_tup[0])].astype(str)
    for pnum in p_tup[1:]:
        choosers[dest_col] = choosers[dest_col] + choosers[add_pn('ptype', pnum)].astype(str)

    # sort the list of ptypes so it is in increasing ptype order, then convert to int
    choosers[dest_col] = choosers[dest_col].apply(lambda x: ''.join(sorted(x))).astype(int)


def hh_choosers(indiv_utils, hhsize):
    """
    Build a chooser table for calculating house utilities for all households of specified hhsize

    The choosers table will have one row per household with columns containing the indiv_utils
    for all non-extra (i.e. cdap_rank <- MAX_HHSIZE) persons. That makes 3 columns for each
    individual. e.g. the utilities of person with cdap_rank 1 will be included as M_p1, N_p1, H_p1

    The chooser table will also contain interaction columns for all possible interactions involving
    from 2 to 3 persons (actually MAX_INTERACTION_CARDINALITY, which is currently 3).

    The interaction columns list the ptypes of the persons in the interaction set, sorted by ptype.
    For instance the interaction between persons with cdap_rank 1 and three and ptypes will
    be listed in a column named 'p1_p3' and for a household where persons p1 and p3 are 2 and 4
    will a row value of 24 in the p1_p3 column.

    Parameters
    ----------
    indiv_utils : pandas.DataFrame
        CDAP utilities for each individual, ignoring interactions.
        ind_utils has index of _persons_index_ and a column for each alternative
        i.e. three columns 'M' (Mandatory), 'N' (NonMandatory), 'H' (Home)

    hhsize : int
        household size for which the choosers table should be built. Households with more than
        MAX_HHSIZE members will be included with MAX_HHSIZE choosers since the are handled the
        same, and the activities of the extra members are assigned afterwards

    Returns
    -------
    choosers : pandas.DataFrame
        choosers households of hhsize with activity utility columns interaction columns
        for all (non-extra) household members
    """

    # we want to merge the ptype and M, N, and H utilities for each individual in the household
    merge_cols = [_hh_id_, _ptype_, 'M', 'N', 'H']

    if hhsize > MAX_HHSIZE:
        raise RuntimeError("hh_choosers hhsize > MAX_HHSIZE")

    if hhsize < MAX_HHSIZE:
        include_households = (indiv_utils[_hh_size_] == hhsize)
    else:
        # we want to include larger households along with MAX_HHSIZE households
        include_households = (indiv_utils[_hh_size_] >= MAX_HHSIZE)

    # start with all the individuals with cdap_rank of 1 (thus there will be one row per household)
    choosers = indiv_utils.loc[include_households & (indiv_utils['cdap_rank'] == 1), merge_cols]
    # rename columns, adding pn suffix (e.g. ptype_p1, M_p1) to all columns except hh_id
    choosers.columns = add_pn(merge_cols, 1)

    # for each of the higher cdap_ranks
    for pnum in range(2, hhsize+1):

        # df with merge columns for indiv with cdap_rank of pnum
        rhs = indiv_utils.loc[include_households & (indiv_utils['cdap_rank'] == pnum), merge_cols]
        # rename columns, adding pn suffix (e.g. ptype_p1, M_p1) to all columns except hh_id
        rhs.columns = add_pn(merge_cols, pnum)

        # merge this cdap_rank into choosers
        choosers = pd.merge(left=choosers, right=rhs, on=_hh_id_)

    # we set index to _hh_id_ choosers has one row per household
    set_hh_index(choosers)

    # coerce utilities to float (merge apparently makes column type objects)
    for pnum in range(1, hhsize+1):
        pn_cols = add_pn(['M', 'N', 'H'], pnum)
        choosers[pn_cols] = choosers[pn_cols].astype(float)

    # add interaction columns for all 2 and 3 person interactions
    for i in range(2, min(hhsize, MAX_INTERACTION_CARDINALITY)+1):
        for tup in itertools.combinations(list(range(1, hhsize+1)), i):
            add_interaction_column(choosers, tup)

    return choosers


def household_activity_choices(indiv_utils, interaction_coefficients, hhsize,
                               trace_hh_id=None, trace_label=None):
    """
    Calculate household utilities for each activity pattern alternative for households of hhsize
    The resulting activity pattern for each household will be coded as a string of activity codes.
    e.g. 'MNHH' for a 4 person household with activities Mandatory, NonMandatory, Home, Home

    Parameters
    ----------
    indiv_utils : pandas.DataFrame
        CDAP utilities for each individual, ignoring interactions
        ind_utils has index of _persons_index_ and a column for each alternative
        i.e. three columns 'M' (Mandatory), 'N' (NonMandatory), 'H' (Home)

    interaction_coefficients : pandas.DataFrame
        Rules and coefficients for generating interaction specs for different household sizes

    hhsize : int
        the size of household for which activity perttern should be calculated (1..MAX_HHSIZE)

    Returns
    -------
    choices : pandas.Series
        the chosen cdap activity pattern for each household represented as a string (e.g. 'MNH')
        with same index (_hh_index_) as utils

    """

    if hhsize == 1:
        # for 1 person households, there are no interactions to account for
        # and the household utils are the same as the individual utils
        choosers = vars = None
        # extract the individual utilities for individuals from hhsize 1 households
        utils = indiv_utils.loc[indiv_utils[_hh_size_] == 1, [_hh_id_, 'M', 'N', 'H']]
        # index on household_id, not person_id
        set_hh_index(utils)
    else:

        choosers = hh_choosers(indiv_utils, hhsize=hhsize)

        spec = build_cdap_spec(interaction_coefficients, hhsize,
                               trace_spec=(trace_hh_id in choosers.index),
                               trace_label=trace_label)

        utils = simulate.eval_utilities(spec, choosers, trace_label=trace_label)

    if len(utils.index) == 0:
        return pd.Series(dtype='float64')

    probs = logit.utils_to_probs(utils, trace_label=trace_label)

    # select an activity pattern alternative for each household based on probability
    # result is a series indexed on _hh_index_ with the (0 based) index of the column from probs
    idx_choices, rands = logit.make_choices(probs, trace_label=trace_label)

    # convert choice expressed as index into alternative name from util column label
    choices = pd.Series(utils.columns[idx_choices].values, index=utils.index)

    if trace_hh_id:

        if hhsize > 1:
            tracing.trace_df(choosers, '%s.hhsize%d_choosers' % (trace_label, hhsize),
                             column_labels=['expression', 'person'])

        tracing.trace_df(utils, '%s.hhsize%d_utils' % (trace_label, hhsize),
                         column_labels=['expression', 'household'])
        tracing.trace_df(probs, '%s.hhsize%d_probs' % (trace_label, hhsize),
                         column_labels=['expression', 'household'])
        tracing.trace_df(choices, '%s.hhsize%d_activity_choices' % (trace_label, hhsize),
                         column_labels=['expression', 'household'])
        tracing.trace_df(rands, '%s.hhsize%d_rands' % (trace_label, hhsize),
                         columns=[None, 'rand'])

    return choices


def unpack_cdap_indiv_activity_choices(persons, hh_choices,
                                       trace_hh_id, trace_label):
    """
    Unpack the household activity choice list into choices for each (non-extra) household member

    Parameters
    ----------
    persons : pandas.DataFrame
        Table of persons data indexed on _persons_index_
        We expect, at least, columns [_hh_id_, 'cdap_rank']
    hh_choices : pandas.Series
        household activity pattern is encoded as a string (of length hhsize) of activity codes
        e.g. 'MNHH' for a 4 person household with activities Mandatory, NonMandatory, Home, Home

    Returns
    -------
    cdap_indiv_activity_choices : pandas.Series
        series contains one activity per individual hh member, indexed on _persons_index_
    """

    cdap_indivs = persons['cdap_rank'] <= MAX_HHSIZE

    indiv_activity = pd.merge(
        left=persons.loc[cdap_indivs, [_hh_id_, 'cdap_rank']],
        right=hh_choices.to_frame(name='hh_choices'),
        left_on=_hh_id_,
        right_index=True
    )

    # resulting dataframe has columns _hh_id_,'cdap_rank', hh_choices indexed on _persons_index_

    indiv_activity['cdap_activity'] = ''

    # for each cdap_rank (1..5)
    for i in range(MAX_HHSIZE):
        pnum_i = (indiv_activity['cdap_rank'] == i+1)
        indiv_activity.loc[pnum_i, ['cdap_activity']] = indiv_activity[pnum_i]['hh_choices'].str[i]

    cdap_indiv_activity_choices = indiv_activity['cdap_activity']

    # if DUMP:
    #     tracing.trace_df(cdap_indiv_activity_choices,
    #                      '%s.DUMP.cdap_indiv_activity_choices' % trace_label,
    #                      transpose=False, slicer='NONE')

    return cdap_indiv_activity_choices


def extra_hh_member_choices(persons, cdap_fixed_relative_proportions, locals_d,
                            trace_hh_id, trace_label):
    """
    Generate the activity choices for the 'extra' household members who weren't handled by cdap

    Following the CTRAMP HouseholdCoordinatedDailyActivityPatternModel, "a separate,
    simple cross-sectional distribution is looked up for the remaining household members"

    The cdap_fixed_relative_proportions spec is handled like an activitysim logit utility spec,
    EXCEPT that the values computed are relative proportions, not utilities
    (i.e. values are not exponentiated before being normalized to probabilities summing to 1.0)

    Parameters
    ----------
    persons : pandas.DataFrame
        Table of persons data indexed on _persons_index_
         We expect, at least, columns [_hh_id_, _ptype_]
    cdap_fixed_relative_proportions
        spec to compute/specify the relative proportions of each activity (M, N, H)
        that should be used to choose activities for additional household members
        not handled by CDAP.
    locals_d : Dict
        dictionary of local variables that eval_variables adds to the environment
        for an evaluation of an expression that begins with @

    Returns
    -------
    choices : pandas.Series
        list of alternatives chosen for all extra members, indexed by _persons_index_
    """

    trace_label = tracing.extend_trace_label(trace_label, 'extra_hh_member_choices')

    # extra household members have cdap_ran > MAX_HHSIZE
    choosers = persons[persons['cdap_rank'] > MAX_HHSIZE]

    if len(choosers.index) == 0:
        return pd.Series(dtype='float64')

    # eval the expression file
    values = simulate.eval_variables(cdap_fixed_relative_proportions.index, choosers, locals_d)

    # cdap_fixed_relative_proportions computes relative proportions by ptype, not utilities
    proportions = values.dot(cdap_fixed_relative_proportions)

    # convert relative proportions to probability
    probs = proportions.div(proportions.sum(axis=1), axis=0)

    # select an activity pattern alternative for each person based on probability
    # idx_choices is a series (indexed on _persons_index_ ) with the chosen alternative represented
    # as the integer (0 based) index of the chosen column from probs
    idx_choices, rands = logit.make_choices(probs, trace_label=trace_label)

    # convert choice from column index to activity name
    choices = pd.Series(probs.columns[idx_choices].values, index=probs.index)

    # if DUMP:
    #     tracing.trace_df(proportions, '%s.DUMP.extra_proportions' % trace_label,
    #                      transpose=False, slicer='NONE')
    #     tracing.trace_df(probs, '%s.DUMP.extra_probs' % trace_label,
    #                      transpose=False, slicer='NONE')
    #     tracing.trace_df(choices, '%s.DUMP.extra_choices' % trace_label,
    #                      transpose=False,
    #                      slicer='NONE')

    if trace_hh_id:
        tracing.trace_df(proportions, '%s.extra_hh_member_choices_proportions' % trace_label,
                         column_labels=['expression', 'person'])
        tracing.trace_df(probs, '%s.extra_hh_member_choices_probs' % trace_label,
                         column_labels=['expression', 'person'])
        tracing.trace_df(choices, '%s.extra_hh_member_choices_choices' % trace_label,
                         column_labels=['expression', 'person'])
        tracing.trace_df(rands, '%s.extra_hh_member_choices_rands' % trace_label,
                         columns=[None, 'rand'])

    return choices


def _run_cdap(
        persons,
        person_type_map,
        cdap_indiv_spec,
        interaction_coefficients,
        cdap_fixed_relative_proportions,
        locals_d,
        trace_hh_id, trace_label):
    """
    Implements core run_cdap functionality on persons df (or chunked subset thereof)
    Aside from chunking of persons df, params are passed through from run_cdap unchanged

    Returns pandas Dataframe with two columns:
        cdap_activity : str
            activity for that person expressed as 'M', 'N', 'H'
        cdap_rank : int
            activities for persons with cdap_rank <= MAX_HHSIZE are determined by cdap
            'extra' household members activities are assigned by cdap_fixed_relative_proportions
    """

    # assign integer cdap_rank to each household member
    # persons with cdap_rank 1..MAX_HHSIZE will be have their activities chose by CDAP model
    # extra household members, will have activities assigned by in fixed proportions
    assign_cdap_rank(persons, person_type_map, trace_hh_id, trace_label)
    chunk.log_df(trace_label, 'persons', persons)

    # Calculate CDAP utilities for each individual, ignoring interactions
    # ind_utils has index of 'person_id' and a column for each alternative
    # i.e. three columns 'M' (Mandatory), 'N' (NonMandatory), 'H' (Home)
    indiv_utils = individual_utilities(persons[persons.cdap_rank <= MAX_HHSIZE],
                                       cdap_indiv_spec, locals_d,
                                       trace_hh_id, trace_label)
    chunk.log_df(trace_label, 'indiv_utils', indiv_utils)

    # compute interaction utilities, probabilities, and hh activity pattern choices
    # for each size household separately in turn up to MAX_HHSIZE
    hh_choices_list = []
    for hhsize in range(1, MAX_HHSIZE+1):

        choices = household_activity_choices(
            indiv_utils, interaction_coefficients, hhsize=hhsize,
            trace_hh_id=trace_hh_id, trace_label=trace_label)

        hh_choices_list.append(choices)

    del indiv_utils
    chunk.log_df(trace_label, 'indiv_utils', None)

    # concat all the household choices into a single series indexed on _hh_index_
    hh_activity_choices = pd.concat(hh_choices_list)
    chunk.log_df(trace_label, 'hh_activity_choices', hh_activity_choices)

    # unpack the household activity choice list into choices for each (non-extra) household member
    # resulting series contains one activity per individual hh member, indexed on _persons_index_
    cdap_person_choices \
        = unpack_cdap_indiv_activity_choices(persons, hh_activity_choices,
                                             trace_hh_id, trace_label)

    # assign activities to extra household members (with cdap_rank > MAX_HHSIZE)
    # resulting series contains one activity per individual hh member, indexed on _persons_index_
    extra_person_choices \
        = extra_hh_member_choices(persons, cdap_fixed_relative_proportions, locals_d,
                                  trace_hh_id, trace_label)

    # concat cdap and extra persoin choices into a single series
    # this series will be the same length as the persons dataframe and be indexed on _persons_index_

    person_choices = pd.concat([cdap_person_choices, extra_person_choices])

    persons['cdap_activity'] = person_choices
    chunk.log_df(trace_label, 'persons', persons)

    # if DUMP:
    #     tracing.trace_df(hh_activity_choices, '%s.DUMP.hh_activity_choices' % trace_label,
    #                      transpose=False, slicer='NONE')
    #     tracing.trace_df(cdap_results, '%s.DUMP.cdap_results' % trace_label,
    #                      transpose=False, slicer='NONE')

    result = persons[['cdap_rank', 'cdap_activity']]

    del persons
    chunk.log_df(trace_label, 'persons', None)

    return result


def cdap_calc_row_size(choosers, cdap_indiv_spec, trace_label):

    sizer = chunk.RowSizeEstimator(trace_label)

    # NOTE we chunk chunk_id
    num_choosers = choosers['chunk_id'].max() + 1
    rows_per_chunk_id = len(choosers) / num_choosers

    chooser_row_size = len(choosers.columns)

    sizer.add_elements(chooser_row_size, 'persons')
    sizer.add_elements(len(cdap_indiv_spec), 'indiv_utils')
    sizer.add_elements(1, 'hh_activity_choices')
    sizer.add_elements(1, 'cdap_rank')
    sizer.add_elements(1, 'cdap_activity')

    row_size = sizer.get_hwm()

    # scale row_size by average number of chooser rows per chunk_id
    row_size = row_size * rows_per_chunk_id

    return row_size


def run_cdap(
        persons,
        person_type_map,
        cdap_indiv_spec,
        cdap_interaction_coefficients,
        cdap_fixed_relative_proportions,
        locals_d,
        chunk_size=0, trace_hh_id=None, trace_label=None):
    """
    Choose individual activity patterns for persons.

    Parameters
    ----------
    persons : pandas.DataFrame
        Table of persons data. Must contain at least a household ID, household size,
        person type category, and age, plus any columns used in cdap_indiv_spec
    cdap_indiv_spec : pandas.DataFrame
        CDAP spec for individuals without taking any interactions into account.
    cdap_interaction_coefficients : pandas.DataFrame
        Rules and coefficients for generating interaction specs for different household sizes
    cdap_fixed_relative_proportions : pandas.DataFrame
        Spec to for the relative proportions of each activity (M, N, H)
        to choose activities for additional household members not handled by CDAP
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
        in either the cdap_indiv_spec or cdap_fixed_relative_proportions expression files
    chunk_size: int
        Chunk size or 0 for no chunking
    trace_hh_id : int
        hh_id to trace or None if no hh tracing
    trace_label : str
        label for tracing or None if no tracing

    Returns
    -------
    choices : pandas.DataFrame

        dataframe is indexed on _persons_index_ and has two columns:

        cdap_activity : str
            activity for that person expressed as 'M', 'N', 'H'
    """

    trace_label = tracing.extend_trace_label(trace_label, 'cdap')

    row_size = chunk_size and cdap_calc_row_size(persons, cdap_indiv_spec, trace_label)

    result_list = []
    # segment by person type and pick the right spec for each person type
    for i, persons_chunk, chunk_trace_label \
            in chunk.adaptive_chunked_choosers_by_chunk_id(persons, chunk_size, row_size, trace_label):

        cdap_results = \
            _run_cdap(persons_chunk,
                      person_type_map,
                      cdap_indiv_spec,
                      cdap_interaction_coefficients,
                      cdap_fixed_relative_proportions,
                      locals_d,
                      trace_hh_id, chunk_trace_label)

        result_list.append(cdap_results)

    # FIXME: this will require 2X RAM
    # if necessary, could append to hdf5 store on disk:
    # http://pandas.pydata.org/pandas-docs/stable/io.html#id2
    if len(result_list) > 1:
        cdap_results = pd.concat(result_list)

    if trace_hh_id:

        tracing.trace_df(cdap_results,
                         label="cdap",
                         columns=['cdap_rank', 'cdap_activity'],
                         warn_if_empty=True)

    # return choices column as series
    return cdap_results['cdap_activity']



import logging

import numpy as np
import pandas as pd

from activitysim.core import chunk
from activitysim.core import config
from activitysim.core import expressions
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core import simulate
from activitysim.core import tracing

from activitysim.abm.models.util.trip import generate_alternative_sizes, get_time_windows
from activitysim.core.interaction_sample_simulate import _interaction_sample_simulate
from activitysim.core.mem import force_garbage_collect


logger = logging.getLogger(__name__)

TOUR_DURATION_COLUMN = 'duration'
NUM_ALTERNATIVES = 'num_alts'
MAIN_LEG_DURATION = 'main_leg_duration'
IB_DURATION = 'inbound_duration'
OB_DURATION = 'outbound_duration'
NUM_OB_STOPS = 'num_outbound_stops'
NUM_IB_STOPS = 'num_inbound_stops'
HAS_OB_STOPS = 'has_outbound_stops'
HAS_IB_STOPS = 'has_inbound_stops'
LAST_OB_STOP = 'last_outbound_stop'
FIRST_IB_STOP = 'last_inbound_stop'

SCHEDULE_ID = 'schedule_id'

OUTBOUND_FLAG = 'outbound'

TEMP_COLS = [NUM_OB_STOPS, LAST_OB_STOP,
             NUM_IB_STOPS, FIRST_IB_STOP,
             NUM_ALTERNATIVES
             ]


def generate_schedule_alternatives(tours):
    """
    For a set of tours, build out the potential schedule alternatives
    for the main leg, outbound leg, and inbound leg. This process handles
    the change in three steps.

    Definitions:
      - Main Leg: The time from last outbound stop to the first inbound stop.
                  If the tour does not include any intermediate stops this
                  will represent the full tour duration.
      - Outbound Leg: The time from the tour origin to the last outbound stop
      - Inbound Leg: The time from the first inbound stop to the tour origin

    1. For tours with no intermediate stops, it simple asserts a main leg
       duration equal to the tour duration.

    2. For tours with an intermediate stop on one of the legs, calculate
       all possible time combinations that are allowed in the duration

    3. For tours with an intermediate stop on both legs, calculate
       all possible time combinations that are allowed in the tour
       duration

    :param tours: pd.Dataframe: Must include a field for tour duration
            and boolean fields indicating intermediate inbound or outbound
            stops.
    :return: pd.Dataframe: Potential time duration windows.
    """
    assert set([NUM_IB_STOPS, NUM_OB_STOPS, TOUR_DURATION_COLUMN]).issubset(tours.columns)

    stop_pattern = tours[HAS_OB_STOPS].astype(int) + tours[HAS_IB_STOPS].astype(int)

    no_stops = no_stops_patterns(tours[stop_pattern == 0])
    one_way = stop_one_way_only_patterns(tours[stop_pattern == 1])
    two_way = stop_two_way_only_patterns(tours[stop_pattern > 1])

    schedules = pd.concat([no_stops, one_way, two_way], sort=True)
    schedules[SCHEDULE_ID] = np.arange(1, schedules.shape[0] + 1)

    return schedules


def no_stops_patterns(tours):
    """
    Asserts the tours with no intermediate stops have a main leg duration equal
    to the tour duration and set inbound and outbound windows equal to zero.
    :param tours: pd.Dataframe: Tours with no intermediate stops.
    :return: pd.Dataframe: Main leg duration, outbound leg duration, and inbound leg duration
    """
    alternatives = tours[[TOUR_DURATION_COLUMN]].rename(columns={TOUR_DURATION_COLUMN: MAIN_LEG_DURATION})
    alternatives[[IB_DURATION, OB_DURATION]] = 0
    return alternatives.astype(int)


def stop_one_way_only_patterns(tours, travel_duration_col=TOUR_DURATION_COLUMN):
    """
    Calculates potential time windows for tours with a single leg with intermediate
    stops. It calculates all possibilities for the main leg and one tour leg to sum to
    the tour duration. The other leg is asserted with a duration of zero.
    :param tours: pd.Dataframe: Tours with no intermediate stops.
    :return: pd.Dataframe: Main leg duration, outbound leg duration, and inbound leg duration
            The return dataframe is indexed to the tour input index
    """
    if tours.empty:
        return None

    assert travel_duration_col in tours.columns

    indexes, patterns, pattern_sizes = get_pattern_index_and_arrays(tours.index, tours[travel_duration_col],
                                                                    one_way=True)
    direction = np.repeat(tours[HAS_OB_STOPS], pattern_sizes)

    inbound = np.where(direction == 0, patterns[:, 1], 0)
    outbound = np.where(direction == 1, patterns[:, 1], 0)

    patterns = pd.DataFrame(index=indexes, data=np.column_stack((patterns[:, 0], outbound, inbound)),
                            columns=[MAIN_LEG_DURATION, OB_DURATION, IB_DURATION])
    patterns.index.name = tours.index.name

    return patterns


def stop_two_way_only_patterns(tours, travel_duration_col=TOUR_DURATION_COLUMN):
    """
    Calculates potential time windows for tours with intermediate stops on both
    legs. It calculates all possibilities for the main leg and both tour legs to
    sum to the tour duration.
    :param tours: pd.Dataframe: Tours with no intermediate stops.
    :return: pd.Dataframe: Main leg duration, outbound leg duration, and inbound leg duration
            The return dataframe is indexed to the tour input index
    """
    if tours.empty:
        return None

    assert travel_duration_col in tours.columns

    indexes, patterns, _ = get_pattern_index_and_arrays(tours.index, tours[travel_duration_col], one_way=False)

    patterns = pd.DataFrame(index=indexes, data=patterns,
                            columns=[MAIN_LEG_DURATION, OB_DURATION, IB_DURATION])
    patterns.index.name = tours.index.name

    return patterns


def trip_schedule_calc_row_size(choosers, trace_label):
    """
    rows_per_chunk calculator for trip_scheduler
    """

    sizer = chunk.RowSizeEstimator(trace_label)

    chooser_row_size = len(choosers.columns)
    spec_columns = 3

    sizer.add_elements(chooser_row_size + spec_columns, 'choosers')

    row_size = sizer.get_hwm()
    return row_size


def get_pattern_index_and_arrays(tour_indexes, durations, one_way=True):
    """
    A helper method to quickly calculate all of the potential time windows
    for a given set of tour indexes and durations.
    :param tour_indexes: List of tour indexes
    :param durations: List of tour durations
    :param one_way: If True, calculate windows for only one tour leg. If False,
                    calculate tour windows for both legs
    :return: np.array: Tour indexes repeated for valid pattern
             np.array: array with a column for main tour leg, outbound leg, and inbound leg
             np.array: array with the number of patterns for each tour
    """
    max_columns = 2 if one_way else 3
    max_duration = np.max(durations)
    time_windows = get_time_windows(max_duration, max_columns)

    patterns = []
    pattern_sizes = []

    for duration in durations:
        possible_windows = time_windows[:max_columns, np.where(time_windows.sum(axis=0) == duration)[0]]
        possible_windows = np.unique(possible_windows, axis=1).transpose()
        patterns.append(possible_windows)
        pattern_sizes.append(possible_windows.shape[0])

    indexes = np.repeat(tour_indexes, pattern_sizes)

    patterns = np.concatenate(patterns)
    # If we've done everything right, the indexes
    # calculated above should be the same length as
    # the pattern options
    assert patterns.shape[0] == len(indexes)

    return indexes, patterns, pattern_sizes


def get_spec_for_segment(model_settings, spec_name, segment):
    """
    Read in the model spec
    :param model_settings: model settings file
    :param spec_name: name of the key in the settings file
    :param segment: which segment of the spec file do you want to read
    :return: array of utility equations
    """

    omnibus_spec = simulate.read_model_spec(file_name=model_settings[spec_name])

    spec = omnibus_spec[[segment]]

    # might as well ignore any spec rows with 0 utility
    spec = spec[spec.iloc[:, 0] != 0]
    assert spec.shape[0] > 0

    return spec


def run_trip_scheduling_choice(spec, tours, skims, locals_dict,
                               chunk_size, trace_hh_id, trace_label):

    NUM_TOUR_LEGS = 3
    trace_label = tracing.extend_trace_label(trace_label, 'interaction_sample_simulate')

    # FIXME: The duration, start, and end should be ints well before we get here...
    tours[TOUR_DURATION_COLUMN] = tours[TOUR_DURATION_COLUMN].astype(np.int8)

    # Setup boolean columns to make it easier to identify
    # intermediate stops later in the model.
    tours[HAS_OB_STOPS] = tours[NUM_OB_STOPS] >= 1
    tours[HAS_IB_STOPS] = tours[NUM_IB_STOPS] >= 1

    # Calculate a matrix with the appropriate alternative sizes
    # based on the total tour duration. This is used to calculate
    # chunk sizes.
    max_duration = tours[TOUR_DURATION_COLUMN].max()
    alt_sizes = generate_alternative_sizes(max_duration, NUM_TOUR_LEGS)

    # Assert the number of tour leg schedule alternatives for each tour
    tours[NUM_ALTERNATIVES] = 1
    tours.loc[tours[HAS_OB_STOPS] != tours[HAS_IB_STOPS], NUM_ALTERNATIVES] = tours[TOUR_DURATION_COLUMN] + 1
    tours.loc[tours[HAS_OB_STOPS] & tours[HAS_IB_STOPS], NUM_ALTERNATIVES] = \
        tours.apply(lambda x: alt_sizes[1, x.duration], axis=1)

    # If no intermediate stops on the tour, then then main leg duration
    # equals the tour duration and the intermediate durations are zero
    tours.loc[~tours[HAS_OB_STOPS] & ~tours[HAS_IB_STOPS], MAIN_LEG_DURATION] = tours[TOUR_DURATION_COLUMN]
    tours.loc[~tours[HAS_OB_STOPS] & ~tours[HAS_IB_STOPS], [IB_DURATION, OB_DURATION]] = 0

    # We only need to determine schedules for tours with intermediate stops
    indirect_tours = tours.loc[tours[HAS_OB_STOPS] | tours[HAS_IB_STOPS]]

    # Crudely calculate a chunk size
    # 5=number of columns in the alternatives (3 leg times + index)
    tour_row_size = (4 + len(tours.columns)) * indirect_tours[NUM_ALTERNATIVES].mean()

    # rpc, effective_chunk_size = chunk.rows_per_chunk(chunk_size, tour_row_size, indirect_tours.shape[0], trace_label)
    row_size = chunk_size and trip_schedule_calc_row_size(indirect_tours, trace_label)
    # Iterate through the chunks
    result_list = []
    for i, choosers, chunk_trace_label in \
            chunk.adaptive_chunked_choosers(indirect_tours, chunk_size, row_size, trace_label):

        # Sort the choosers and get the schedule alternatives
        choosers = choosers.sort_index()
        schedules = generate_schedule_alternatives(choosers).sort_index()

        # Assuming we did the max_alt_size calculation correctly,
        # we should get the same sizes here.
        assert choosers[NUM_ALTERNATIVES].sum() == schedules.shape[0]

        # Run the simulation
        choices = _interaction_sample_simulate(
            choosers=choosers,
            alternatives=schedules,
            spec=spec,
            choice_column=SCHEDULE_ID,
            allow_zero_probs=True, zero_prob_choice_val=-999,
            want_logsums=False,
            skims=skims,
            locals_d=locals_dict,
            trace_label=chunk_trace_label,
            trace_choice_name='trip_schedule_stage_1',
            estimator=None
        )

        assert len(choices.index) == len(choosers.index)

        choices = schedules[schedules[SCHEDULE_ID].isin(choices)].drop(columns='tour_id')

        result_list.append(choices)

        force_garbage_collect()

    # FIXME: this will require 2X RAM
    # if necessary, could append to hdf5 store on disk:
    # http://pandas.pydata.org/pandas-docs/stable/io.html#id2
    if len(result_list) > 1:
        choices = pd.concat(result_list)

    assert len(choices.index) == len(indirect_tours.index)

    # The choices here are only the indirect tours, so the durations
    # need to be updated on the main tour dataframe.
    tours.update(choices[[MAIN_LEG_DURATION, OB_DURATION, IB_DURATION]])

    # Cleanup data types and drop temporary columns
    tours[[MAIN_LEG_DURATION, OB_DURATION, IB_DURATION]] = \
        tours[[MAIN_LEG_DURATION, OB_DURATION, IB_DURATION]].astype(np.int8)
    tours = tours.drop(columns=TEMP_COLS)

    return tours


@inject.step()
def trip_scheduling_choice(
        trips,
        tours,
        skim_dict,
        chunk_size,
        trace_hh_id):

    trace_label = 'trip_scheduling_choice'
    model_settings = config.read_model_settings('trip_scheduling_choice.yaml')
    spec = get_spec_for_segment(model_settings, 'SPECIFICATION', 'stage_one')

    trips_df = trips.to_frame()
    tours_df = tours.to_frame()

    outbound_trips = trips_df[trips_df[OUTBOUND_FLAG]]
    inbound_trips = trips_df[~trips_df[OUTBOUND_FLAG]]

    last_outbound_trip = trips_df.loc[outbound_trips.groupby('tour_id')['trip_num'].idxmax()]
    first_inbound_trip = trips_df.loc[inbound_trips.groupby('tour_id')['trip_num'].idxmin()]

    tours_df[NUM_OB_STOPS] = outbound_trips.groupby('tour_id').size().reindex(tours.index) - 1
    tours_df[NUM_IB_STOPS] = inbound_trips.groupby('tour_id').size().reindex(tours.index) - 1
    tours_df[LAST_OB_STOP] = last_outbound_trip[['tour_id', 'origin']].set_index('tour_id').reindex(tours.index)
    tours_df[FIRST_IB_STOP] = first_inbound_trip[['tour_id', 'destination']].set_index('tour_id').reindex(tours.index)

    preprocessor_settings = model_settings.get('PREPROCESSOR', None)

    if preprocessor_settings:
        # hack: preprocessor adds origin column in place if it does not exist already
        od_skim_stack_wrapper = skim_dict.wrap('origin', 'destination')
        do_skim_stack_wrapper = skim_dict.wrap('destination', 'origin')
        obib_skim_stack_wrapper = skim_dict.wrap(LAST_OB_STOP, FIRST_IB_STOP)

        skims = [od_skim_stack_wrapper, do_skim_stack_wrapper, obib_skim_stack_wrapper]

        locals_dict = {
            "od_skims": od_skim_stack_wrapper,
            "do_skims": do_skim_stack_wrapper,
            "obib_skims": obib_skim_stack_wrapper
        }

        simulate.set_skim_wrapper_targets(tours_df, skims)

        expressions.assign_columns(
            df=tours_df,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label)

    tours_df = run_trip_scheduling_choice(spec, tours_df, skims, locals_dict, chunk_size, trace_hh_id, trace_label)

    pipeline.replace_table("tours", tours_df)




import numpy as np
import pandas as pd
import pytest

import activitysim.abm.models.trip_departure_choice as tdc
from activitysim.abm.models.util.trip import get_time_windows
from activitysim.core import los
from .test_pipeline import setup_dirs


@pytest.fixture(scope='module')
def trips():
    outbound_array = [True, True, False, False, False, True, True, False, False, True]

    trips = pd.DataFrame(data={'tour_id': [1, 1, 2, 2, 2, 2, 2, 3, 3, 4],
                               'trip_duration': [2, 2, 7, 7, 7, 12, 12, 4, 4, 5],
                               'inbound_duration': [0, 0, 7, 7, 7, 0, 0, 4, 4, 5],
                               'main_leg_duration': [4, 4, 2, 2, 2, 2, 2, 1, 1, 2],
                               'outbound_duration': [2, 2, 0, 0, 0, 12, 12, 0, 0, 5],
                               'trip_count': [2, 2, 3, 3, 3, 2, 2, 2, 2, 1],
                               'trip_num': [1, 2, 1, 2, 3, 1, 2, 1, 2, 1],
                               'outbound': outbound_array,
                               'chunk_id': [1, 1, 2, 2, 2, 2, 2, 3, 3, 4],
                               'is_work': [True, True, False, False, False, False, False, False, False, True],
                               'is_school': [False, False, False, False, False, False, False, True, True, False],
                               'is_eatout': [False, False, True, True, True, True, True, False, False, False],
                               'start': [8, 8, 18, 18, 18, 18, 18, 24, 24, 19],
                               'end': [14, 14, 39, 39, 39, 39, 39, 29, 29, 26],
                               'origin': [3, 5, 15, 12, 24, 8, 17, 8, 9, 6],
                               'destination': [5, 9, 12, 24, 20, 17, 18, 9, 11, 14],
                               }, index=range(10))

    trips.index.name = 'trip_id'
    return trips


@pytest.fixture(scope='module')
def settings():
    return {"skims_file": "skims.omx",
            "skim_time_periods": {
                "labels": ['EA', 'AM', 'MD', 'PM', 'NT']}
            }


@pytest.fixture(scope='module')
def model_spec():
    index = ["@(df['stop_time_duration'] * df['is_work'].astype(int)).astype(int)",
             "@(df['stop_time_duration'] * df['is_school'].astype(int)).astype(int)",
             "@(df['stop_time_duration'] * df['is_eatout'].astype(int)).astype(int)"]

    values = {'inbound': [0.933020, 0.370260, 0.994840],
              'outbound': [0.933020, 0.370260, 0.994840]
              }

    return pd.DataFrame(index=index, data=values)


def test_build_patterns(trips):
    time_windows = get_time_windows(48, 3)
    patterns = tdc.build_patterns(trips, time_windows)
    patterns = patterns.sort_values(['tour_id', 'outbound', 'trip_num'])

    assert patterns.shape[0] == 34
    assert patterns.shape[1] == 6
    assert patterns.index.name == tdc.TOUR_LEG_ID

    output_columns = [tdc.TOUR_ID, tdc.PATTERN_ID, tdc.TRIP_NUM,
                      tdc.STOP_TIME_DURATION, tdc.TOUR_ID, tdc.OUTBOUND]

    assert set(output_columns).issubset(patterns.columns)


def test_get_tour_legs(trips):
    tour_legs = tdc.get_tour_legs(trips)
    assert tour_legs.index.name == tdc.TOUR_LEG_ID
    assert np.unique(tour_legs[tdc.TOUR_ID].values).shape[0] == np.unique(trips[tdc.TOUR_ID].values).shape[0]


def test_generate_alternative(trips):
    alts = tdc.generate_alternatives(trips, tdc.STOP_TIME_DURATION)
    assert alts.shape[0] == 67
    assert alts.shape[1] == 1

    assert alts.index.name == tdc.TRIP_ID
    assert alts.columns[0] == tdc.STOP_TIME_DURATION

    pd.testing.assert_series_equal(trips.groupby(trips.index)['trip_duration'].max(),
                                   alts.groupby(alts.index)[tdc.STOP_TIME_DURATION].max(),
                                   check_names=False)


def test_apply_stage_two_model(model_spec, trips):
    setup_dirs()
    departures = tdc.apply_stage_two_model(model_spec, trips, 0, 'TEST Trip Departure')
    assert len(departures) == len(trips)
    pd.testing.assert_index_equal(departures.index, trips.index)

    departures = pd.concat([trips, departures], axis=1)



import pandas as pd
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core import tracing


@inject.step()
def step1():

    table1 = pd.DataFrame({'c': [1, 2, 3]})
    inject.add_table('table1', table1)


@inject.step()
def step2():

    table1 = pd.DataFrame({'c': [2, 4, 6]})
    inject.add_table('table2', table1)


@inject.step()
def step3():

    table1 = pd.DataFrame({'c': [3, 6, 9]})
    inject.add_table('table3', table1)


@inject.step()
def step_add_col():

    table_name = inject.get_step_arg('table_name')
    assert table_name is not None

    col_name = inject.get_step_arg('column_name')
    assert col_name is not None

    table = pipeline.get_table(table_name)

    assert col_name not in table.columns

    table[col_name] = table.index + (1000 * len(table.columns))

    pipeline.replace_table(table_name, table)


@inject.step()
def step_forget_tab():

    table_name = inject.get_step_arg('table_name')
    assert table_name is not None

    table = pipeline.get_table(table_name)

    pipeline.drop_table(table_name)


@inject.step()
def create_households(trace_hh_id):

    df = pd.DataFrame({'household_id': [1, 2, 3], 'home_zone_id': {100, 100, 101}})
    inject.add_table('households', df)

    pipeline.get_rn_generator().add_channel('households', df)

    tracing.register_traceable_table('households', df)




# ActivitySim
# See full license in LICENSE.txt.
import time
import datetime
import psutil
import logging
import gc


from activitysim.core import config
from activitysim.core import inject

logger = logging.getLogger(__name__)

MEM = {}
HWM = {}
DEFAULT_TICK_LEN = 30


def force_garbage_collect():
    was_disabled = not gc.isenabled()
    if was_disabled:
        gc.enable()
    gc.collect()
    if was_disabled:
        gc.disable()


def GB(bytes):
    gb = (bytes / (1024 * 1024 * 1024.0))
    return round(gb, 2)


def init_trace(tick_len=None, file_name="mem.csv", write_header=False):
    MEM['tick'] = 0
    if file_name is not None:
        MEM['file_name'] = file_name
    if tick_len is None:
        MEM['tick_len'] = DEFAULT_TICK_LEN
    else:
        MEM['tick_len'] = tick_len

    logger.info("init_trace file_name %s" % file_name)

    # - check for optional process name prefix
    MEM['prefix'] = inject.get_injectable('log_file_prefix', 'main')

    if write_header:
        with config.open_log_file(file_name, 'w') as log_file:
            print("process,time,rss,used,available,percent,event", file=log_file)


def trace_hwm(tag, value, timestamp, label):

    hwm = HWM.setdefault(tag, {})

    if value > hwm.get('mark', 0):
        hwm['mark'] = value
        hwm['timestamp'] = timestamp
        hwm['label'] = label


def log_hwm():

    for tag in HWM:
        hwm = HWM[tag]
        logger.info("high water mark %s: %.2f timestamp: %s label: %s" %
                    (tag, hwm['mark'], hwm['timestamp'], hwm['label']))

    with config.open_log_file(MEM['file_name'], 'a') as log_file:
        for tag in HWM:
            hwm = HWM[tag]
            print("%s high water mark %s: %.2f timestamp: %s label: %s" %
                  (MEM['prefix'], tag, hwm['mark'], hwm['timestamp'], hwm['label']), file=log_file)


def trace_memory_info(event=''):

    if not MEM:
        return

    last_tick = MEM['tick']
    tick_len = MEM['tick_len'] or float('inf')

    t = time.time()
    if (t - last_tick < tick_len) and not event:
        return

    force_garbage_collect()

    vmi = psutil.virtual_memory()

    MEM['tick'] = t

    current_process = psutil.Process()
    rss = current_process.memory_info().rss
    for child in current_process.children(recursive=True):
        try:
            rss += child.memory_info().rss
        except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
            pass

    timestamp = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")

    trace_hwm('rss', GB(rss), timestamp, event)
    trace_hwm('used', GB(vmi.used), timestamp, event)

    if event:
        logger.info(f"trace_memory_info {event} rss: {GB(rss)}GB used: {GB(vmi.used)} GB percent: {vmi.percent}%")

    with config.open_log_file(MEM['file_name'], 'a') as output_file:

        print("%s, %s, %.2f, %.2f, %.2f, %s%%, %s" %
              (MEM['prefix'],
               timestamp,
               GB(rss),
               GB(vmi.used),
               GB(vmi.available),
               vmi.percent,
               event), file=output_file)


def get_rss():

    mi = psutil.Process().memory_info()

    # cur_mem = mi.vms
    cur_mem = mi.rss

    return cur_mem



# ActivitySim
# See full license in LICENSE.txt.
import os

import pandas as pd
import pandas.testing as pdt

from activitysim.core import pipeline
from activitysim.core import inject
from activitysim.core import mp_tasks

from test_multi_zone import example_path
from test_multi_zone import mtc_example_path
from test_multi_zone import setup_dirs
from test_multi_zone import regress_3_zone

# set the max households for all tests (this is to limit memory use on travis)
HOUSEHOLDS_SAMPLE_SIZE = 25

# household with WALK_TRANSIT tours and trips
HH_ID_3_ZONE = 2848373


def test_mp_run():

    configs_dir = [example_path('configs_3_zone'), mtc_example_path('configs')]
    data_dir = example_path('data_3')

    setup_dirs(configs_dir, data_dir)
    inject.add_injectable('settings_file_name', 'settings_mp.yaml')
    inject.add_injectable('households_sample_size', HOUSEHOLDS_SAMPLE_SIZE)
    inject.add_injectable('trace_hh_id', HH_ID_3_ZONE)

    run_list = mp_tasks.get_run_list()
    mp_tasks.print_run_list(run_list)

    # do this after config.handle_standard_args, as command line args may override injectables
    injectables = ['data_dir', 'configs_dir', 'output_dir', 'settings_file_name',
                   'households_sample_size', 'trace_hh_id']
    injectables = {k: inject.get_injectable(k) for k in injectables}

    mp_tasks.run_multiprocess(run_list, injectables)
    pipeline.open_pipeline('_')
    regress_3_zone()
    pipeline.close_pipeline()


if __name__ == '__main__':

    test_mp_run()
